"""

Testkit command line tool caller.

"""
import logging
import subprocess
from typing import List


logger = logging.getLogger('metta_testkit:testkit')

TESTKITCLIENT_CLI_CONFIG_FILE_DEFAULT = './testkit.yml'
""" Testkit config configuration file key """

TESTKITCLIENT_WORKING_DIR_DEFAULT = '.'
""" Testkit Client default working dir """
TESTKITCLIENT_BIN_PATH = '/home/james/Documents/Mirantis/tools/testkit/testkit'
""" Testkit bin exec for the subprocess """


class TestkitClient:
    """Shell client for interacting with the testkit bin."""

    def __init__(self, config_file: str = TESTKITCLIENT_CLI_CONFIG_FILE_DEFAULT,
                 working_dir: str = TESTKITCLIENT_WORKING_DIR_DEFAULT,
                 debug: bool = False):
        """Initialize Testkit command executer.

        Parameters:
        -----------
        config_file (str) : path to testkit config file, typically
            testkit.yml

        debug (bool) : passed to the testkit client to tell it to enable
            verbose debugging output.

        """
        self.config_file = config_file
        """ Path to config file """
        self.working_dir = working_dir
        """ Python subprocess working dir to execute testkit in.
            This may be relevant in cases where ssh keys have a relative path """

        self.debug = debug
        """ should testkit be run with verbose output enabled ? """

        self.bin = TESTKITCLIENT_BIN_PATH
        """ path to testkit executable """

    def version(self):
        """Return testkit client version."""
        return self._run(['version'], return_output=True)

    def create(self, opts: List[str]):
        """Run the testkit create command."""
        return self._run(['create'] + opts)

    def system_rm(self, system_name: str):
        """Remove a system from testkit."""
        return self._run(['system', 'rm', system_name])

    # this syntax makes it easier to read
    # pylint: disable=inconsistent-return-statements
    def _run(self, args: List[str], return_output=False):
        """Run a testkit command.

        Parameters:
        -----------
        args (List[str]) : arguments to pass to the testkit bin using
            subprocess

        """
        cmd = [self.bin, f'--file={self.config_file}']

        if self.debug:
            cmd += ['--debug']

        cmd += args

        if not return_output:
            logger.error("running testkit command: %s", " ".join(cmd))
            res = subprocess.run(cmd, cwd=self.working_dir, check=True, text=True, check=True)
            res.check_returncode()
        else:
            logger.debug("running testkit command with output capture: %s", " ".join(cmd))
            res = subprocess.run(cmd, cwd=self.working_dir, shell=False, stdout=subprocess.PIPE,
                                 check=True)
            res.check_returncode()
            return res.stdout.decode('utf-8')



"""

Kubernetes workload plugin for creating infra from a multi-doc yaml file.

This is possible, but not very clean for management of resources.  The python
kubernetes api offers a helper for creating resources, but not easy approach
for removing the resources.

"""

import logging
from typing import List, Any, Dict

import yaml

from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta.workload import WorkloadBase, WorkloadInstanceBase

from .kubeapi_client import KubernetesApiClientPlugin, METTA_PLUGIN_ID_KUBERNETES_CLIENT

logger = logging.getLogger('metta.contrib.kubernetes.workload.yaml')

KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL = 'kubernetes'
KUBERNETES_YAML_WORKLOAD_CONFIG_BASE = 'workload.yaml'

KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_NAMESPACE = "namespace"
KUBERNETES_YAML_WORKLOAD_CONFIG_DEFAULT_NAMESPACE = "default"
KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_FILE = "file"
KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_YAML = "yaml"

METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD = 'metta_kubernetes_yaml'
""" workload plugin_id for the metta_kubernetes yaml plugin """


class KubernetesYamlWorkloadPlugin(WorkloadBase):
    """Metta workload plugin for Kubernetes workloads created from yaml."""

    def __init__(self, environment, instance_id,
                 label: str = KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL,
                 base: Any = KUBERNETES_YAML_WORKLOAD_CONFIG_BASE):
        """Run the super constructor but also set class properties.

        This implements the args part of the client interface.

        Here we expect to receive a path to a KUBECONFIG file with a context
        set and we create a Kubernetes client for use.  After that this can
        provide Core api clients as per the kubernetes SDK

        Parameters:
        -----------
        config_file (str): String path to the kubernetes config file to use

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """Create a workload instance from a set of fixtures.

        Parameters:
        -----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """
        try:
            client = fixtures.get_plugin(
                plugin_type=METTA_PLUGIN_TYPE_CLIENT,
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)
        except KeyError as err:
            raise NotImplementedError(
                "Workload could not find the needed client: "
                f"{METTA_PLUGIN_ID_KUBERNETES_CLIENT}") from err

        workload_config = self.environment.config.load(self.config_label)

        namespace = workload_config.get(
            [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_NAMESPACE],
            default=KUBERNETES_YAML_WORKLOAD_CONFIG_DEFAULT_NAMESPACE)

        # YAML config needs to come from a yaml file path or inline yaml config
        resource_yaml = workload_config.get(
            [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_YAML], default=[])
        file = workload_config.get(
            [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_FILE], default=None)

        if yaml is None and file is None:
            raise ValueError(
                "Either inline yaml or a file path to a yaml is required.")

        return KubernetesYamlWorkloadInstance(
            client, namespace, data=resource_yaml, file=file)

    def info(self):
        """Return dict data about this plugin for introspection."""
        workload_config = self.environment.config.load(self.config_label)

        return {
            'workload': {
                'deployment': {
                    'namespace': workload_config.get(
                        [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_NAMESPACE],
                        default=KUBERNETES_YAML_WORKLOAD_CONFIG_DEFAULT_NAMESPACE),
                    'yaml': workload_config.get(
                        [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_YAML],
                        default='None'),
                    'file': workload_config.get(
                        [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_FILE],
                        default='None')
                },
                'required_fixtures': {
                    'kubernetes': {
                        'plugin_type': METTA_PLUGIN_TYPE_CLIENT,
                        'plugin_id': 'metta_kubernetes'
                    }
                }
            }
        }


# @TODO create a remove method as well.
# pylint: disable=too-few-public-methods
class KubernetesYamlWorkloadInstance(WorkloadInstanceBase):
    """Instance of the k8s yaml workload, used to manage a particular run.

    Parameters:
    -----------
    client (kubeapi_client) : metta kubernetes kubeapi_client object

    namespace (str) : string namespace to use for all craeted resources

    data (Dict[str, Any]) : laoded yaml content to be passed to the kubeapi client

    OR

    file (str) : file path to a file containing the yaml.

    """

    def __init__(self, client: KubernetesApiClientPlugin, namespace: str,
                 data: Dict[str, Any], file: str):
        """Comfigure new workload instance."""
        self.client = client
        self.namespace = namespace
        self.data = data
        self.file = file

        self.k8s_objects: List[object] = []

    def apply(self):
        """Use the passed yaml to create k8s resources.

        Returns:
        --------
        Any created resources.

        """
        if self.file:
            with open(self.file) as res_file:
                resources_yaml = yaml.safe_load_all(res_file)

                for resource in resources_yaml:
                    self.k8s_objects.append(self.client.utils_create_from_dict(
                        data=resource,
                        namespace=self.namespace
                    ))

        else:
            self.k8s_objects.append(self.client.utils_create_from_dict(
                data=self.data,
                namespace=self.namespace
            ))

        return self.k8s_objects

    # @TODO write the docker container rm
    def destroy(self):
        """Destroy a created docker run."""



"""

Metta Terraform.

metta contrib functionality for terraform.  Primarily a terraform provisioner
plugin.

"""
from typing import Any

from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER
from mirantis.testing.metta_cli.base import METTA_PLUGIN_TYPE_CLI

from .cli import TerraformCliPlugin, METTA_TERRAFORM_CLI_PLUGIN_ID
from .provisioner import (TerraformProvisionerPlugin, TERRAFORM_PROVISIONER_CONFIG_LABEL,
                          METTA_TERRAFORM_PROVISIONER_PLUGIN_ID)


@Factory(plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID)
def metta_plugin_factory_provisioner_terraform(environment: Environment, instance_id: str = "",
                                               label: str = TERRAFORM_PROVISIONER_CONFIG_LABEL,
                                               base: Any = LOADED_KEY_ROOT):
    """Create an metta provisionersss dict plugin."""
    return TerraformProvisionerPlugin(environment, instance_id, label, base)


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_TERRAFORM_CLI_PLUGIN_ID)
def metta_terraform_factory_cli_terraform(environment: Environment, instance_id: str = ''):
    """Create an info cli plugin."""
    return TerraformCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added.

    """



"""

Metta CLI : Ansible Provisioner commands.

Comamnds for itneracting with the ansible plugins, primarily the provisioner

"""
import logging

from mirantis.testing.metta.environment import Environment

from mirantis.testing.metta_cli.base import CliBase, cli_output
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER

from .provisioner import METTA_ANSIBLE_PROVISIONER_PLUGIN_ID

logger = logging.getLogger('metta.cli.ansible')

METTA_ANSIBLE_CLI_PLUGIN_ID = 'metta_ansible'
""" cli plugin_id for the info plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class AnsibleCliPlugin(CliBase):
    """Fire command/group generator for Ansible commands."""

    def fire(self):
        """Return a dict of commands."""
        if self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_PROVISIONER,
                plugin_id=METTA_ANSIBLE_PROVISIONER_PLUGIN_ID,
                exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'ansible': AnsibleGroup(self.environment)
                }
            }

        return {}


class AnsibleGroup():
    """Base Fire command group for Ansible commands."""

    def __init__(self, environment: Environment):
        """Inject environment into command gorup."""
        self.environment = environment

    def _select_provisioner(self, instance_id: str = ''):
        """Pick a matching provisioner."""
        if instance_id:
            return self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_PROVISIONER,
                plugin_id=METTA_ANSIBLE_CLI_PLUGIN_ID, instance_id=instance_id)

        # Get the highest priority provisioner
        return self.environment.fixtures.get(
            plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id=METTA_ANSIBLE_CLI_PLUGIN_ID)

    def info(self, provisioner: str = ''):
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin

        info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority
            }
        }
        info.update(plugin.info())

        return cli_output(info)

    def prepare(self, provisioner: str = ''):
        """Run provisioner prepare."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.prepare()

    def apply(self, provisioner: str = ''):
        """Run provisioner apply."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.apply()

    def destroy(self, provisioner: str = ''):
        """Run provisioner destroy."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.destroy()



"""

Terraform clie handler.

Here are is all of the functionality which actually runs terraform commands
on the command line using subprocess.

"""

import logging
import json
import os
import time
import subprocess
from typing import Dict, List

logger = logging.getLogger('metta_terraform:client')

TERRAFORM_CLIENT_DEFAULT_BINARY = 'terraform'
""" default terraform executable for subprocess """


class TerraformClient:
    """Shell client for running terraform using subprocess."""

    # this is what it takes to configure the terraform client
    # pylint: disable=too-many-arguments
    def __init__(self, working_dir: str, state_path: str, vars_path: str,
                 variables: Dict[str, str], binary: str = TERRAFORM_CLIENT_DEFAULT_BINARY):
        """Initialize Terraform client.

        Parameters:
        -----------
        working_dir (str) : string path to the python where the terraform root
            module/plan is, so that subprocess/tf can use that as a pwd

        state_path (str) : path to where the terraform state should be kept

        vars_path (str) : string path to where the vars file should be written.

        variables (Dict[str,str]) : terraform variables dict which will be
            written to a vars file.

        """
        self.vars = variables
        self.working_dir = working_dir
        self.state_path = state_path
        self.vars_path = vars_path

        self.terraform_bin = binary

    def state(self):
        """Return the terraform state contents."""
        try:
            with open(os.path.join(self.working_dir, 'terraform.tfstate')) as json_file:
                return json.load(json_file)
        except FileNotFoundError:
            logger.debug("Terraform client found no state file")
            return None

    def init(self):
        """Run terraform init.

        init is something that can be run once for a number of jobs in parallel
        we lock the process. If a lock is in place, then we just wait for an
        unlock and return.
        Other terraform actions lock themselves, and we want to fail if the
        operation is locked, but here we just want to skip it.

        """
        try:
            lockfile = os.path.join(os.path.dirname(self.state_path),
                                    '.terraform.metta_mirantis.init.lock')
            if os.path.exists(lockfile):
                logger.info(
                    "terraform .init lock file found.  Skipping init, but waiting for it to finish")
                time_to_wait = 120
                time_counter = 0
                while not os.path.exists(lockfile):
                    time.sleep(5)
                    time_counter += 5
                    if time_counter > time_to_wait:
                        raise BlockingIOError("Timed out when waiting for init lock to go away")
            else:
                os.makedirs(os.path.dirname(os.path.abspath(lockfile)), exist_ok=True)
                with open(lockfile, 'w') as lockfile_object:
                    lockfile_object.write(f"{os.getpid()} is running init")
                try:
                    self._run(['init'], with_vars=False, with_state=False)
                finally:
                    os.remove(lockfile)
        except subprocess.CalledProcessError as err:
            logger.error("Terraform client failed to run init in %s: %s", self.working_dir,
                         err.output)
            raise Exception("Terraform client failed to run init") from err

    def apply(self):
        """Apply a terraform plan."""
        try:
            self._run(['apply', '-auto-approve'], with_state=True,
                      with_vars=True, return_output=False)
        except subprocess.CalledProcessError as err:
            logger.error("Terraform client failed to run apply in %s: %s", self.working_dir,
                         err.stderr)
            raise Exception("Terraform client failed to run") from err

    def plan(self):
        """Check a terraform plan."""
        try:
            self._run(['plan'], with_state=True,
                      with_vars=True, return_output=False)
        except subprocess.CalledProcessError as err:
            logger.error("Terraform client failed to run plan in %s: %s", self.working_dir,
                         err.stderr)
            raise Exception("Terraform client failed to plan") from err

    def destroy(self):
        """Apply a terraform plan."""
        try:
            self._run(['destroy', '-auto-approve'], with_state=True,
                      with_vars=True, return_output=False)
        except subprocess.CalledProcessError as err:
            logger.error("Terraform client failed to run init in %s: %s", self.working_dir,
                         err.output)
            raise Exception("Terraform client failed to run destroy") from err

    def output(self, name: str = ''):
        """Retrieve terraform outputs.

        Run the terraform output command, to retrieve outputs.
        Outputs are returned always as json as it is the only way to machine
        parse outputs properly.

        Returns:
        --------
        If you provided a name, then a single output is returned, otherwise a
        dict of outputs is returned.

        """
        args = ['output', '-json']
        """ collect subprocess args to pass """

        try:
            if name:
                output = self._run(
                    args, [name], with_vars=False, return_output=True)
            else:
                output = self._run(args, with_vars=False, return_output=True)
        except subprocess.CalledProcessError as err:
            logger.error("Terraform client failed to run init in %s: %s", self.working_dir,
                         err.output)
            raise Exception("Terraform client failed to retrieve output") from err

        return json.loads(output)

    def _make_vars_file(self):
        """Write the vars file."""
        vars_path = self.vars_path

        try:
            os.makedirs(
                os.path.dirname(
                    os.path.abspath(vars_path)),
                exist_ok=True)
            with open(vars_path, 'w') as var_file:
                json.dump(self.vars, var_file, sort_keys=True, indent=4)
        except Exception as err:
            raise Exception(f"Could not create terraform vars file: {vars_path}") from err

    def _run(self, args: List[str], append_args: List[str] = None, with_state=True,
             with_vars=True, return_output=False):
        """Run terraform CLI command."""
        cmd = [self.terraform_bin]
        cmd += args

        if with_vars:
            self._make_vars_file()
            cmd += [f'-var-file={self.vars_path}']
        if with_state:
            cmd += [f'-state={self.state_path}']

        if append_args is not None:
            cmd += append_args

        # improve readability
        # pylint: disable=no-else-return
        if not return_output:
            logger.debug("running terraform command: %s", " ".join(cmd))
            res = subprocess.run(cmd, cwd=self.working_dir, check=True, text=True)
            res.check_returncode()
            return res
        else:
            logger.debug(
                "running terraform command with output capture: %s",
                " ".join(cmd))
            res = subprocess.run(cmd, cwd=self.working_dir, shell=False, check=True,
                                 stdout=subprocess.PIPE)
            res.check_returncode()
            return res.stdout.decode('utf-8')



"""

Metta client plugin for executing commands (ssh/winrm) on hosts provisioned.

On a cluster of hosts provisioned using the Launchpad provisioner, execute
a command over ssh/winrm on one of the hosts.

"""
import logging
from typing import List

from mirantis.testing.metta.environment import Environment

from .launchpad import LaunchpadClient

logger = logging.getLogger('metta.contrib.kubernetes.client')


METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID = "metta_launchpad_exec"
""" metta plugin ID for the metta exec client plugin """


class LaunchpadExecClientPlugin:
    """Client for exec into hosts using launchpad."""

    def __init__(self, environment: Environment,
                 instance_id: str, client: LaunchpadClient):
        """Create launchpad exec plugin.

        Parameters:
        -----------
        client (LaunchpadClient) : A configured launchpad client to use to run
            exec commands

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        self.client = client
        """ Launchpad client """

    def hosts(self, deep: bool = False):
        """List the hosts in the cluster."""
        config = self.client.describe_config()

        if deep:
            host_list = list(config['spec']['hosts'])
        else:
            host_list = []
            for host in config['spec']['hosts']:
                list_host = {
                    'role': host['role']
                }
                if 'ssh' in host:
                    list_host.update({
                        'is_windows': False,
                        'address': host['ssh']['address']
                    })
                if 'winrm' in host:
                    list_host.update({
                        'is_windows': True,
                        'address': host['winrm']['address']
                    })

                host_list.append(list_host)

        return host_list

    def exec(self, cmds: List[str], host_index: int):
        """Exec a string arg list command on a single host."""
        return self.client.exec(host_index, cmds)



"""

Launchpad metta provisioner plugin.

Provisioner/Install  Mirantis products onto an existing cluster using
Launchpad.

"""
import os.path
import logging
from typing import Any, List, Dict

import yaml

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
from configerus.validator import ValidationError

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.provisioner import ProvisionerBase
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta_docker import METTA_PLUGIN_ID_DOCKER_CLIENT
from mirantis.testing.metta_kubernetes import METTA_PLUGIN_ID_KUBERNETES_CLIENT
from mirantis.testing.metta_mirantis import (METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
                                             METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID)

from .launchpad import LaunchpadClient, METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT
from .exec_client import METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID

logger = logging.getLogger('mirantis.testing.metta.provisioner:launchpad')

METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID = "metta_launchpad"
""" Metta plugin_id for the launchpad provisioner plugin """

METTA_LAUNCHPAD_CONFIG_LABEL = 'launchpad'
""" Launchpad config label for configuration """
METTA_LAUNCHPAD_CONFIG_ROOT_PATH_KEY = 'root.path'
""" config key for a base file path that should be used for any relative paths """
METTA_LAUNCHPAD_CONFIG_KEY = 'config'
""" which config key will provide the launchpad yml """
METTA_LAUNCHPAD_CONFIG_API_ACCESSPOINT_KEY = 'api.accesspoint'
""" config key for the API endpoint, usually the manager load-balancer """
METTA_LAUNCHPAD_CONFIG_API_USERNAME_KEY = 'api.username'
""" config key for the API username """
METTA_LAUNCHPAD_CONFIG_API_PASSWORD_KEY = 'api.password'
""" config key for the API password """
METTA_LAUNCHPAD_CONFIG_HOSTS_KEY = 'config.spec.hosts'
""" config key for the list of hosts as per the launchpad spec """
METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY = 'config_file'
""" Launchpad config cli key to tell us where to put the launchpad yml file """
METTA_LAUNCHPAD_CLI_WORKING_DIR_KEY = 'working_dir'
""" Launchpad config cli configuration working dir key """
METTA_LAUNCHPAD_CLI_CLUSTEROVERRIDE_KEY = 'cluster_name'
""" If provided, this config key will override a cluster name pulled from yaml"""
METTA_LAUNCHPAD_CLI_ACCEPTLICENSE_KEY = 'cli.accept-license'
""" If provided, this config key will tell the cli to accept the license"""
METTA_LAUNCHPAD_CLI_DISABLETELEMETRY_KEY = 'cli.disable-telemetry'
""" If provided, this config key will tell the cli to disable telemetry"""
METTA_LAUNCHPAD_CLI_DISABLETELEMETRYUPGRADECHECK_KEY = 'cli.disable-upgrade-check'
""" If provided, this config key will tell the cli to disable upgrade checks"""
METTA_LAUNCHPAD_CLI_CONFIG_DOCKER_VERSION_DEFAULT = '1.40'
""" Default value for the docker client version number."""

METTA_LAUNCHPAD_VALIDATE_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        'config': {
            'type': ['object', 'null'],
        },
        'api': {
            'type': 'object',
            'properties': {
                'endpoint': {'type': ['string', 'null']}
            },
        },
        'cli': {
            'accept-license': {'type': 'bool'},
            'disable-telemetry': {'type': 'bool'},
        },
        'cluster_name': {
            'type': 'string'
        },
        'config_file': {
            'type': 'string'
        },
        'working_dir': {
            'type': 'string'
        }
    },
    'required': ['config_file', 'config', 'api']
}
""" Validation jsonschema for terraform config contents """
METTA_LAUNCHPAD_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: METTA_LAUNCHPAD_VALIDATE_JSONSCHEMA
}
""" configerus validation target to match validate Launchpad config """

METTA_LAUNCHPAD_CONFIG_VALIDATE_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        'apiVersion': {'type': 'string'},
        'kind': {'type': 'string'},
        'hosts': {
            'type': 'array',
            'items': {
                'type': 'object'
            }
        },
        'spec': {
            'type': 'object',
            'properties': {
                'mcr': {'type': 'object'},
                'mke': {'type': 'object'},
                'msr': {'type': 'object'}
            },
            'required': ['mcr', 'mke']
        }
    },
    'required': ['apiVersion', 'kind', 'spec']
}
""" Validation jsonschema for launchpad configuration """
METTA_LAUNCHPAD_CONFIG_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: METTA_LAUNCHPAD_CONFIG_VALIDATE_JSONSCHEMA
}
""" configerus jsonschema validation target for launchpad config """


# pylint: disable=too-many-instance-attributes
class LaunchpadProvisionerPlugin(ProvisionerBase):
    """Launchpad provisioner class.

    Use this to provision a system using Mirantis launchpad

    """

    def __init__(self, environment: Environment, instance_id: str,
                 label: str = METTA_LAUNCHPAD_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
        """Configure a new Launchpad provisioner plugin instance."""
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        self.fixtures = Fixtures()
        """ keep a collection of fixtures that this provisioner creates """

        self.downloaded_bundle_users: List[str] = []
        """ track user bundles that have been downloaded to avoid unecessary repeats """

        self.config_label: str = label
        self.config_base: str = base

        """ load all of the launchpad configuration """
        launchpad_config_loaded = self.environment.config.load(label)

        # Run confgerus validation on the config using our above defined
        # jsonschema
        try:
            launchpad_config_loaded.get(base, validator=METTA_LAUNCHPAD_VALIDATE_TARGET)
        except ValidationError as err:
            raise ValueError("Launchpad config failed validation.") from err

        working_dir = launchpad_config_loaded.get(
            [base, METTA_LAUNCHPAD_CLI_WORKING_DIR_KEY])
        """ if launchpad needs to be run in a certain path, set it with this config """
        if not os.path.isabs(working_dir):
            # did a relative path root get passed in as config?
            root_path = self.backend_output_name = launchpad_config_loaded.get(
                [base, METTA_LAUNCHPAD_CONFIG_ROOT_PATH_KEY])
            if root_path:
                working_dir = os.path.join(root_path, working_dir)
            working_dir = os.path.abspath(working_dir)

        # decide on a path for the runtime launchpad.yml file
        self.config_file: str = launchpad_config_loaded.get(
            [base, METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY],
            default=METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT)
        if not os.path.isabs(self.config_file):
            # A relative [ath for the config file is expected to be relative to
            # the working dir]
            self.config_file = os.path.abspath(
                os.path.join(working_dir, self.config_file))

        cluster_name_override = launchpad_config_loaded.get(
            [base, METTA_LAUNCHPAD_CLI_CLUSTEROVERRIDE_KEY], default='')
        """ Can hardcode the cluster name if it can't be take from the yaml file """

        accept_license = launchpad_config_loaded.get(
            [base, METTA_LAUNCHPAD_CLI_ACCEPTLICENSE_KEY], default=True)
        """ should the client accept the license """
        disable_telemetry = launchpad_config_loaded.get(
            [base, METTA_LAUNCHPAD_CLI_DISABLETELEMETRY_KEY], default=True)
        """ should the client disable telemetry """
        disable_upgrade_check = launchpad_config_loaded.get(
            [base, METTA_LAUNCHPAD_CLI_DISABLETELEMETRYUPGRADECHECK_KEY], default=True)
        """ should the client disable upgrade checks """

        logger.debug("Creating Launchpad API client")
        self.client = LaunchpadClient(
            config_file=self.config_file,
            working_dir=working_dir,
            cluster_name_override=cluster_name_override,
            accept_license=accept_license,
            disable_telemetry=disable_telemetry,
            disable_upgrade_check=disable_upgrade_check)

        # If we can, it makes sense to build the docker and k8s client fixtures now.
        # This will only be possible in cases where we have an existing config file
        # and we can download a client bundle.
        # We try that here, even though it is verbose and ugly, so that we have
        # the clients available for introspection.
        # We probably shouldn't, but it allows some flexibility.
        if os.path.exists(self.config_file):
            try:
                self._make_fixtures()

            # We don't want this constructor to fail on a functionality test no matter the cause
            # pylint: disable=broad-except
            except BaseException:
                logger.debug("Could not create launchpad plugins on bootstrap.")
                # we most likely failed because we don't have enough info get
                # make fixtures from, as launchpad hasn't installed yet

    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Get info about a provisioner plugin.

        Returns:
        --------
        Dict of introspective information about this plugin_info

        """
        plugin = self
        client = self.client

        loaded = self.environment.config.load(self.config_label)

        info = {
            'plugin': {
                'config_label': plugin.config_label,
                'config_base': plugin.config_base,
                'downloaded_bundle_users': plugin.downloaded_bundle_users
            },
            'client': {
                'cluster_name_override': client.cluster_name_override,
                'config_file': client.config_file,
                'working_dir': client.working_dir,
                'bin': client.bin
            },
            'config': {
                'contents': loaded.get(self.config_base),
            }
        }

        if deep:
            try:
                info['config']['interpreted'] = client.describe_config()
                info['bundles'] = {user: client.bundle(
                    user) for user in client.bundle_users()}

            # pylint: disable=broad-except
            except Exception:
                # There are many legitimate cases where this fails
                pass

            fixtures = {}
            for fixture in self.fixtures:
                fixture_info = {
                    'fixture': {
                        'plugin_type': fixture.plugin_type,
                        'plugin_id': fixture.plugin_id,
                        'instance_id': fixture.instance_id,
                        'priority': fixture.priority,
                    }
                }
                if hasattr(fixture.plugin, 'info'):
                    plugin_info = fixture.plugin.info()
                    if isinstance(plugin_info, dict):
                        fixture_info.update(plugin_info)
                fixtures[fixture.instance_id] = fixture_info
            info['fixtures'] = fixtures

        user = 'admin'
        info['helper'] = {
            'commands': {
                'apply': f"{client.bin} apply -c {client.config_file}",
                'client-config': f"{client.bin} client-config -c {client.config_file} {user}"
            }
        }

        return info

    # pylint: disable=no-self-use
    def prepare(self):
        """Prepare the provisioning cluster for install.

        We ignore this.

        """
        logger.info("Running Launchpad Prepare().  Launchpad has no prepare stage.")

    def apply(self):
        """Bring a cluster up.

        We assume that the cluster is running and the we can pull the required
        yaml from an output fixture in the environment.

        This plugin needs an output fixture, probably of dict type.  It will
        Pull that structure for the launchpad yaml config file and dump it into
        its config path.
        The provisioner can find an output directly from the environment, or
        from a specific fixture source.  If you want the output to come from
        only a specific backend fixture then make sure that a "backend" config
        exists, otherwise just use an "output" config.

        Raises:
        -------
        ValueError if the object has been configured (prepare) with config that
            doesn't work, or if the backend doesn't give valid yml

        Exception if launchpad fails.

        """
        try:
            logger.info("Using launchpad to install products onto backend cluster")
            self._write_launchpad_file()
            self.client.apply()
        except Exception as err:
            raise Exception("Launchpad failed to install") from err

        # Rebuild the fixture list now that we have installed
        self._make_fixtures(reload=True)

    def destroy(self):
        """Ask the client to remove installed resources."""
        self.client.reset()

    # ----- CLUSTER INTERACTION -----

    def _write_launchpad_file(self):
        """Write the config state to the launchpad file."""
        try:
            # load all of the launchpad configuration, force a reload to get up to date contents
            launchpad_config = self.environment.config.load(self.config_label, force_reload=True)
            launchpad_config = launchpad_config.get(
                [self.config_base, METTA_LAUNCHPAD_CONFIG_KEY],
                validator=METTA_LAUNCHPAD_CONFIG_VALIDATE_TARGET)
        except KeyError as err:
            raise ValueError("Could not find launchpad configuration from config.") from err
        except ValidationError as err:
            raise ValueError("Launchpad config failed validation") from err

        # write the launchpad output to our yaml file target (after creating
        # the path)
        os.makedirs(os.path.dirname(os.path.realpath(self.config_file)), exist_ok=True)
        with open(os.path.realpath(self.config_file), 'w') as file:
            yaml.dump(launchpad_config, file)

    # @TODO break this into many make client functions
    # pylint: disable=too-many-locals
    def _make_fixtures(self, user: str = 'admin', reload: bool = False) -> Fixtures:
        """Build fixtures for all of the clients.

        Returns:
        --------
        Fixtures collection of fixtures that have been created

        """
        # get fresh values for the launchpad config (in case it has changed)
        launchpad_config = self.environment.config.load(
            self.config_label, force_reload=reload)

        # holds retrieved bundle information from MKE for all client configs.
        bundle_info = self._mke_client_bundle(user, reload)

        # KUBE Client

        kube_config = os.path.join(bundle_info['path'], 'kube.yml')
        if not os.path.exists(kube_config):
            raise NotImplementedError(
                "Launchpad was asked for a kubernetes client, but not kube config file was in the "
                "client bundle.  Are you sure this is a kube cluster?")

        instance_id = f"{self.instance_id}-{METTA_PLUGIN_ID_KUBERNETES_CLIENT}-{user}"
        fixture = self.environment.add_fixture(
            METTA_PLUGIN_TYPE_CLIENT,
            plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT,
            instance_id=instance_id,
            priority=70,
            arguments={'kube_config_file': kube_config})
        # use the parent UCCTFixturesPlugin methods for adding fixtures
        self.fixtures.add(fixture)

        # DOCKER CLIENT
        #
        # @NOTE we pass in a docker API constraint because I ran into a case where the
        #   client failed because the python library was ahead in API version

        try:
            host = bundle_info['Endpoints']['docker']['Host']
            cert_path = bundle_info['tls_paths']['docker']
        except TypeError as err:
            logger.error(
                "Could not read client bundle properly: %s",
                bundle_info['Endpoints']['docker']['Host'])
            raise err

        instance_id = f"{self.instance_id}-{METTA_PLUGIN_ID_DOCKER_CLIENT}-{user}"
        fixture = self.environment.add_fixture(
            METTA_PLUGIN_TYPE_CLIENT,
            plugin_id=METTA_PLUGIN_ID_DOCKER_CLIENT,
            instance_id=instance_id,
            priority=70,
            arguments={'host': host, 'cert_path': cert_path,
                       'version': METTA_LAUNCHPAD_CLI_CONFIG_DOCKER_VERSION_DEFAULT})
        # use the parent UCCTFixturesPlugin methods for adding fixtures
        self.fixtures.add(fixture)

        # EXEC CLIENT
        #
        instance_id = f"{self.instance_id}-{METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID}-{user}"
        fixture = self.environment.add_fixture(
            METTA_PLUGIN_TYPE_CLIENT,
            plugin_id=METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID,
            instance_id=instance_id,
            priority=70,
            arguments={'client': self.client})
        # use the parent UCCTFixturesPlugin methods for adding fixtures
        self.fixtures.add(fixture)

        # The following clients both need the LB, username and password

        api_accesspoint = launchpad_config.get(
            [self.config_base, METTA_LAUNCHPAD_CONFIG_API_ACCESSPOINT_KEY])
        api_username = launchpad_config.get(
            [self.config_base, METTA_LAUNCHPAD_CONFIG_API_USERNAME_KEY])
        api_password = launchpad_config.get(
            [self.config_base, METTA_LAUNCHPAD_CONFIG_API_PASSWORD_KEY])

        hosts = launchpad_config.get(
            [self.config_base, METTA_LAUNCHPAD_CONFIG_HOSTS_KEY], default=[])
        """ isolate the list of hosts so that we can separate them into roles """

        # MKE Client
        #
        instance_id = f"{self.instance_id}-{METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID}-{user}"
        mke_hosts = [host for host in hosts if host['role'] in ['manager']]
        if len(mke_hosts) > 0:
            fixture = self.environment.add_fixture(
                METTA_PLUGIN_TYPE_CLIENT,
                plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
                instance_id=instance_id,
                priority=70,
                arguments={'accesspoint': api_accesspoint, 'username': api_username,
                           'password': api_password, 'hosts': mke_hosts})
            # use the parent UCCTFixturesPlugin methods for adding fixtures
            self.fixtures.add(fixture)

        # MSR Client
        #
        instance_id = f"{self.instance_id}-{METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID}-{user}"
        msr_hosts = [host for host in hosts if host['role'] in ['msr']]
        if len(msr_hosts) > 0:
            fixture = self.environment.add_fixture(
                METTA_PLUGIN_TYPE_CLIENT,
                plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID,
                instance_id=instance_id,
                priority=70,
                arguments={'accesspoint': None, 'username': api_username,
                           'password': api_password, 'hosts': msr_hosts})
            # use the parent UCCTFixturesPlugin methods for adding fixtures
            self.fixtures.add(fixture)

    def _mke_client_bundle(self, user: str, reload: bool = False):
        """Retrieve the MKE Client bundle metadata using the client."""
        assert self.client, "Don't have a launchpad client configured yet"
        return self.client.bundle(user, reload)



"""

DISCOVER: Functionality for discovering METTA environments.

Discover is respondible for discovering project roots and initial project
configuration, along with related configuration, dependency injections and
fixtures.

"""
import os.path
import sys
import logging
import importlib
from typing import List

from configerus.plugin import Type as ConfigerusType
from configerus.config import Config
from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH, CONFIGERUS_PATH_KEY
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT, CONFIGERUS_DICT_DATA_KEY
from configerus.contrib.env import (PLUGIN_ID_SOURCE_ENV_SPECIFIC, CONFIGERUS_ENV_SPECIFIC_BASE_KEY,
                                    PLUGIN_ID_SOURCE_ENV_JSON, CONFIGERUS_ENV_JSON_ENV_KEY)

from .plugin import METTA_PLUGIN_CONFIG_KEY_PLUGINID, METTA_PLUGIN_CONFIG_KEY_PRIORITY

logger = logging.getLogger('metta.discover')

METTA_CONFIG_LABEL = 'metta'
""" this config label is used as a default for discovery from config """
METTA_CONFIG_SOURCES_KEY = 'config.sources'
""" This config key is used to find additional config sources to add """
METTA_CONFIG_IMPORTS_KEY = 'imports'
""" This config key is used to find modules to import """

METTA_ROOT_FILES = [
    'metta.yml',
    'metta.json'
]
""" The presence of any of these files could be used to mark a conf source """

DEFAULT_SOURCE_PRIORITY = 40
""" Default priority for configerus path sources for root marker files """
DEFAULT_SOURCE_CONFIG_PRIORITY = 70
""" Default priority for configerus sources found in config """

DEFAULT_ENVIRONMENT_ROOT_CONFIG_IF_NO_ROOT_IS_FOUND = {
    'metta': {
        'project': {
            'name': 'none'
        }
    },
    'environments': {
        'missing': {

        }
    }
}
""" If no root is found then this config is passed back as a config source """


def discover_project_root(config: Config, start_path: str,
                          marker_files: List[str] = None):
    """Find a project root path.

    We start looking in the start_path for certain marker files, and if we
    don't find any then we check the parent, recursively. If we never find a
    marker file then we assume that the current path is the root.

    Whichever path we think is the root, we add a a configerus path source.
    That means that you can put whatever you want there for config.

    Commonly, one used just the metta.yml|json file to direct metta to include
    additional paths (such a ./config) from the metta.yml path.

    """
    if marker_files is None:
        marker_files = METTA_ROOT_FILES

    # standardize the path to rid of issues that could be cause by path tricks
    # such as  'path/..'
    start_path = os.path.realpath(start_path)

    # Try to any a path from start+path and up that contains a marker file as a
    # config source.
    check_path = start_path
    depth = 0
    while check_path:
        # if we are in a system /root path then stop scanning
        # we do this by checking if the path's parent is the
        # same as itself.
        if check_path == os.path.dirname(check_path):
            break

        for root_file in marker_files:
            root_file_path = os.path.join(check_path, root_file)

            if os.path.isfile(root_file_path):
                # If we found a marker file in a path, then we add that path as
                # a config source.  If the path contains a ./config then we add
                # that as well.

                if check_path not in sys.path:
                    sys.path.append(check_path)

                priority = DEFAULT_SOURCE_PRIORITY - depth
                if depth:
                    instance_id = f"project-{depth}"
                else:
                    instance_id = "project"
                config.add_source(
                    plugin_id=PLUGIN_ID_SOURCE_PATH,
                    instance_id=instance_id,
                    priority=priority).set_path(check_path)
                logger.info("Added project path as config: %s => %s", check_path, instance_id)

                depth += 1
                break

        # move up one directory and try again
        check_path = os.path.dirname(check_path)

    try:
        config.plugins.get_plugins(type=ConfigerusType.SOURCE)
    except KeyError:
        instance_id = "environment_none"
        logger.warning("No project config found, creating a dummy: %s", instance_id)
        source = config.add_source(
            instance_id='notfound',
            plugin_id=PLUGIN_ID_SOURCE_DICT,
            priority=config.default_priority())
        source.set_data(DEFAULT_ENVIRONMENT_ROOT_CONFIG_IF_NO_ROOT_IS_FOUND)

    return config


def discover_sources_from_config(
        config: Config, label: str = METTA_CONFIG_LABEL, base: str = METTA_CONFIG_SOURCES_KEY):
    """Discover more config sources by loading and processing config.

    Run this is you want to add config sources to a config object as defined
    in config itself

    Parameters:
    -----------
    config (Config) : config object to scan, and to add sources to

    label (str) : config label to load to search for sources
    base (str) : config key that should contain the list of sources

    """
    metta_config = config.load(label)

    config_sources = metta_config.get(base, default={})
    for instance_id in config_sources.keys():
        instance_base = [base, instance_id]

        plugin_id = metta_config.get([instance_base, METTA_PLUGIN_CONFIG_KEY_PLUGINID])
        priority = metta_config.get([instance_base,
                                     METTA_PLUGIN_CONFIG_KEY_PRIORITY],
                                    default=DEFAULT_SOURCE_CONFIG_PRIORITY)

        logger.info("Adding metta sourced config plugin: %s:%s", plugin_id, instance_id)
        plugin = config.add_source(
            plugin_id=plugin_id,
            instance_id=instance_id,
            priority=priority)

        if plugin_id == PLUGIN_ID_SOURCE_PATH:
            source_path = metta_config.get(
                [instance_base, CONFIGERUS_PATH_KEY])
            plugin.set_path(path=source_path)
        elif plugin_id == PLUGIN_ID_SOURCE_DICT:
            source_data = metta_config.get(
                [instance_base, CONFIGERUS_DICT_DATA_KEY])
            plugin.set_data(data=source_data)
        elif plugin_id == PLUGIN_ID_SOURCE_ENV_SPECIFIC:
            source_base = metta_config.get(
                [instance_base, CONFIGERUS_ENV_SPECIFIC_BASE_KEY])
            plugin.set_base(base=source_base)
        elif plugin_id == PLUGIN_ID_SOURCE_ENV_JSON:
            source_env = metta_config.get(
                [instance_base, CONFIGERUS_ENV_JSON_ENV_KEY])
            plugin.set_env(env=source_env)
        elif hasattr(plugin, 'set_data'):
            data = metta_config.get([instance_base, 'data'])
            plugin.set_data(data=data)
        else:
            logger.warning("had no way of configuring new source plugin.")


def discover_imports(config: Config, label: str = METTA_CONFIG_LABEL,
                     base: str = METTA_CONFIG_IMPORTS_KEY):
    """Look in config for module imports.

    Use this if you want to dynamically import some modules defined in config.
    This can be used to load custom plugins that are decorated by the plugin
    Factory, but can include any module by path.

    @NOTE Modules are loaded by path, without a local namespace.

    Parameters:
    -----------
    config (Config) : config object to scan, and to add sources to

    label (str) : config label to load to search for sources
    base (str) : config key that should contain the list of sources

    """
    metta_config = config.load(label)

    imports_config = metta_config.get(base, default={})

    for import_name in imports_config:
        module_path = metta_config.get([base, import_name, CONFIGERUS_PATH_KEY])

        if os.path.isdir(module_path):
            module_path_dir = os.path.dirname(module_path)
            module_path_basename = os.path.basename(module_path)
            if not module_path_basename == import_name:
                logger.warning(
                    "Metta discovery importer cannot import a package (folder) using a"
                    "name other than the folder name: %s != %s",
                    module_path_basename, import_name)
            if module_path_dir not in sys.path:
                sys.path.append(module_path_dir)
            importlib.import_module(module_path_basename)
            logger.debug("Loaded package: %s : %s", module_path_basename, module_path)

        elif os.path.isfile(module_path):
            spec = importlib.util.spec_from_file_location(import_name, module_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            logger.debug("Loaded module: %s : %s", import_name, module_path)

        else:
            raise ValueError(
                f"Could not import requested metta import {import_name} : {module_path}")



"""

An environment is meant to represent a testable cluster as a testing harness.

In METTA an environment is a single configerus Config object and a set of
METTA plugins in a manageable set.

METTA uses the Fixtures object to manage a set of plugins, which allows us to
mix a bunch of plugin objects of different types together and manage them.

@NOTE states were just recently added and are currently inelegant and inefficient
  but should get a refactor soon.

"""
import logging
import random
import string
from typing import List, Dict, Any, Union
from importlib import metadata

from configerus.config import Config
from configerus.loaded import Loaded, LOADED_KEY_ROOT
from configerus.validator import ValidationError
from configerus.contrib.jsonschema import PLUGIN_ID_VALIDATE_JSONSCHEMA
from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH, CONFIGERUS_PATH_KEY
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT, CONFIGERUS_DICT_DATA_KEY
from configerus.contrib.env import (
    PLUGIN_ID_SOURCE_ENV_SPECIFIC, CONFIGERUS_ENV_SPECIFIC_BASE_KEY, PLUGIN_ID_SOURCE_ENV_JSON,
    CONFIGERUS_ENV_JSON_ENV_KEY)

from .plugin import (
    Factory, METTA_PLUGIN_CONFIG_KEY_PLUGINTYPE, METTA_PLUGIN_CONFIG_KEY_PLUGINID,
    METTA_PLUGIN_CONFIG_KEY_INSTANCEID, METTA_PLUGIN_CONFIG_KEY_ARGUMENTS,
    METTA_PLUGIN_CONFIG_KEY_PRIORITY, METTA_PLUGIN_CONFIG_KEY_VALIDATORS)
from .fixtures import (
    Fixtures, Fixture, METTA_FIXTURES_CONFIG_FIXTURES_LABEL, METTA_FIXTURE_VALIDATION_JSONSCHEMA)

logger = logging.getLogger('metta.environment')

DEFAULT_PLUGIN_PRIORITY = 70
""" Default plugin priority when turned into a fixture """

FIXTURE_VALIDATION_TARGET_FORMAT_STRING = 'jsonschema:{key}'
""" python string .format template for string jsonchema configerus formatter definition """

METTA_BOOTSTRAP_ENTRYPOINT = 'metta.bootstrap'
""" SetupTools entry_point used for METTA bootstrap """

METTA_PLUGIN_CONFIG_LABEL_ENVIRONMENTS = 'environments'
""" config label discover a list of environments in a loaded config """
METTA_PLUGIN_CONFIG_KEY_ENVIRONMENTS = 'environments'
""" this key could be used to discover a list of environments in a loaded config """
METTA_PLUGIN_CONFIG_KEY_BOOTSTRAPS_METTA = 'bootstraps.metta'
""" config key for metta bootstraps inside any config block """
METTA_PLUGIN_CONFIG_KEY_ENVIRONMENT_STATES = 'states.available'
""" this config key inside an environment config can describe states """
METTA_PLUGIN_CONFIG_KEY_ENVIRONMENT_DEFAULT_STATE = 'states.default'
""" this config key inside an environment config that overrides using the first state as default """
METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG = 'from_config'
""" config key that indicates that the plugin will be build from aconfig label/key pair """

METTA_ENVIRONMENT_STATE_UNUSED_NAME = 'default'
""" what an environment will use for state name if no states are used in the environment """

DEFAULT_SOURCE_PRIORITY = 40
""" If the environment constructor finds config sources to add, this is their default priority """


# pylint: disable=R0902
class Environment:
    """A testing environment, usually composed of a config and plugins."""

    # pylint: disable=too-many-arguments
    def __init__(self, name: str, config: Config, bootstraps: List[str] = None,
                 config_label: str = '', config_base: Union[str, List[Any]] = None,
                 default_priority: int = DEFAULT_SOURCE_PRIORITY):
        """Initialize environment state.

        Parameters:
        -----------
        config (configerus.Config) : A single config object which will be used
            to define the environment.

        bootstraps (List[str]) : a list of metta bootstraps to run. This will be
            combined with a default set, and may also be combined with a set
            from config.

        default_priority (int) : integer values are used to sort created fixtures.
            this value is used in cases where no priority has been specified.

        # Config Context : the following two parameters will tell the Environment
            object to examine config for additional actions to take.  For example
            additional config source may be added, and fixtures may be created.

        Label (str) : Config label to load to find environment config.
        Base (Mixed) : config base key to use to find environment config.  This can
            be either a string or any nested combination of Lists of strings, which
            Configerus joins into a flat list.

        """
        if bootstraps is None:
            bootstraps = []
        if config_base is None:
            config_base = LOADED_KEY_ROOT

        self.name = name
        """ what does the environment call itself """

        # make a copy of the config object as it will likely be shared across contexts
        # that we don't want to have impact this environment obect
        self.config = config.copy()
        """ Config object that defines the environment """
        self.config_backup = config.copy()
        """ Make a dupe of the starting config which we will use whenever we change state """

        # Config label & base are configerus concepts used for loading data:
        # label (str) : can roughly be compared with config file name, without the file suffix
        #     but it will contain merged config from multiple sources.
        # base (str of List of stuff) : directs the path down the loaded config from the label.
        self.config_label: str = config_label
        """ configerus load label """
        self.config_base = config_base
        """ environment config base """

        self.default_priority = int(default_priority)
        """ Default priority int value used for when no priority is assigned """

        self.fixtures = Fixtures()
        """ fixtures/plugins that can interact with the environment """
        self.default_plugin_priority = DEFAULT_PLUGIN_PRIORITY
        """ Default integer priority for new fixtures """

        self.bootstraps = bootstraps
        """ keep the original list of bootstraps to be applied on every state change """
        self.bootstrapped: List[str] = []
        """ list of bootstraps that have already been applied to prevent repetition """

        self.states: List[str] = []
        """ list of allowed states for the environment """
        self.state: str = ''
        """ currently active state for the environment """

        if not self.config_label:
            # this environment does not have a related configuration to program
            # itself with, but it could have had bootstraps.
            logger.info("New environment created: %s (not from config)", name)

            self.bootstrap(bootstraps)
            # this was the original mechanisms for defining environments, and  does have usecases
            # left today for simple environment definition, but any serious environment usage would
            # be much better served by using the configuration options; it lets you define more
            # config sources, fixtures, bootstraps etc.

        else:

            logger.info("New environment created: %s", name)

            # There is a config dict to add to the environment
            config_environment = self.config.load(self.config_label)

            try:
                config_environment.get(self.config_base)
            except ValueError as err:
                raise ValueError("Environment config not found") from err

            # Grab all of the environment state keys/names. Default to an empty
            # list,
            self.states = list(config_environment.get(
                [self.config_base, METTA_PLUGIN_CONFIG_KEY_ENVIRONMENT_STATES], default={}).keys())
            # If this environment doesn't use states then pretend there is a
            # state named after our default value.
            if len(self.states) == 0:
                state = METTA_ENVIRONMENT_STATE_UNUSED_NAME
            else:
                # if there is no config for default state, then None is passed to the state loader
                # which will ignore all state configurations.
                state = config_environment.get(
                    [config_base, METTA_PLUGIN_CONFIG_KEY_ENVIRONMENT_DEFAULT_STATE], default='')
                """ what state should be activated on construction.  Can be empty """

            # We run this function even if we have no state, just so we can put
            # all of the config loading in that loader.
            self.set_state(state)

    def set_state(self, state: str = METTA_ENVIRONMENT_STATE_UNUSED_NAME):
        """Set the enivronment state to one of the options for the state.

        Reload the environment with a new state.  This reloads the entire
        environment, configuring the environment for one of the preconfigured
        states.

        @NOTE When state is set, all fixtures are forgotten from the environment,
        but if you have a fixture in scope and use it, it is still aware of its
        environment and can still cause change in the scope.  It is up to the
        consumer to be aware of this.

        Parameters:
        -----------
        state (str) : string state label, which indicates that configuration
            from that state set should be included.

        @NOTE not all environments have multiple states, but all environments
        will use this method to initialize.  To accomodate this a None state is
        allowed.

        """
        # If we are already in that requested state then back out
        if self.state == state:
            return

        # is the requested state in the list of avialabler states
        if not (state == METTA_ENVIRONMENT_STATE_UNUSED_NAME or state in self.states):
            raise KeyError(f"Requested environment state has not been configured: {state}")

        self.state = state
        self.config = self.config_backup.copy()

        self.fixtures = Fixtures()
        """ fixtures/plugins that can interact with the environment """
        self.bootstrapped = []
        """ list of bootstraps that have already been applied to prevent repetition """

        # load config for the environment
        config_environment = self.config.load(self.config_label)

        config_base = self.config_base
        """ common config base for all environment states """

        # here we check if there is a state that should come from config.
        # if the state looks liek a real key then we build a config base for it
        # for loading config from that path, otherwsie we leave it as None
        # which is used for testing later in this method.
        if state in [METTA_ENVIRONMENT_STATE_UNUSED_NAME, state]:
            state_config_base = None
        else:
            state_config_base = [config_base, METTA_PLUGIN_CONFIG_KEY_ENVIRONMENT_STATES, state]
            """ config base for the selected state """

        # Check for config sources from the environment configuration that is
        # shared across all states
        self._add_config_sources_from_config(
            label=self.config_label, base=[
                config_base, 'config.sources'])
        # check for state configuration sources
        if state_config_base is not None:
            self._add_config_sources_from_config(
                label=self.config_label, base=[
                    state_config_base, 'config.sources'])

        # Check to see if we should pass any bootstraps to the env factory.
        bootstraps = self.bootstraps.copy()
        bootstraps += config_environment.get(
            [config_base, METTA_PLUGIN_CONFIG_KEY_BOOTSTRAPS_METTA], default=[])
        if state_config_base is not None:
            # Add any bootstraps declared in the env config
            bootstraps += config_environment.get(
                [state_config_base, METTA_PLUGIN_CONFIG_KEY_BOOTSTRAPS_METTA], default=[])

        self.bootstrap(bootstraps)

        # Look for fixture definitions in the environment config.
        #
        # One of two options is available here, either:
        # 1. your environment config has a fixtures dict of fixture definitions, or
        # 2. your environment config has a from_config dict which tells us to look
        #    elsewhere.
        # (the same rules are then applied globally and to the active state)

        # If your environment config has a fixtures definition with from_config
        # then the fixtures will be loaded from a different config source.
        # we will look for config like .fixtures and
        # .fictures.from_config.[label|base]
        if config_environment.has([config_base,
                                   METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
                                   METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG]):
            # here is case #2 ^ - we have been told to look elsewhere for fixtures.
            label = config_environment.get([config_base,
                                            METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
                                            METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG,
                                            'label'],
                                           default=METTA_FIXTURES_CONFIG_FIXTURES_LABEL)
            base = config_environment.get([config_base,
                                           METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
                                           METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG,
                                           'base'],
                                          default=LOADED_KEY_ROOT)
            self.add_fixtures_from_config(label=label, base=base)
        elif config_environment.has([config_base, METTA_FIXTURES_CONFIG_FIXTURES_LABEL]):
            self.add_fixtures_from_config(
                label=label, base=[config_base, METTA_FIXTURES_CONFIG_FIXTURES_LABEL])

        if state_config_base is not None:
            # if your state definition has a fixtures "from_config" section,
            # then it will be loaded from a different config source as describe
            # using a label/base pair
            if config_environment.has([state_config_base,
                                       METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
                                       METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG]):
                label = config_environment.get(
                    [state_config_base,
                     METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
                     METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG,
                     'label'],
                    default=METTA_FIXTURES_CONFIG_FIXTURES_LABEL)
                base = config_environment.get(
                    [state_config_base,
                     METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
                     METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG,
                     'base'],
                    default=LOADED_KEY_ROOT)
                self.add_fixtures_from_config(label=label, base=base)

            # If you state config has inline fixtures then they are added.
            elif config_environment.has([state_config_base, METTA_FIXTURES_CONFIG_FIXTURES_LABEL]):
                self.add_fixtures_from_config(
                    label=label, base=[
                        config_base, METTA_FIXTURES_CONFIG_FIXTURES_LABEL])

    def _add_config_sources_from_config(self, label: str, base: Union[str, List[Any]]):
        """Ddd more config sources based on in config settings.

        Read some config which will tell us where more config can be found.  This lets us use config
        to extend config, and is what make metta entirely extensible from a single metta.yml file.

        In the trade-off battle between configurable and convention, this leans heavily towards
        configuration, but it easily lends to standards.

        Parameters:
        -----------
        label (str) : configurus load label.
        base (str|List[str]) : configerus get key as a base for retrieving all config settings.

        """
        config_environment = self.config.load(label)
        config_sources = config_environment.get(base, default={})

        for instance_id in config_sources.keys():
            instance_base = [base, instance_id]

            # Keep in mind that the following plugin metadata is about
            # configerus plugins, not metta plugins.

            plugin_id = config_environment.get([instance_base, 'plugin_id'])
            priority = config_environment.get(
                [instance_base, 'priority'], default=DEFAULT_SOURCE_PRIORITY)

            logger.debug("Adding metta sourced config plugin to '%s' environment: %s:%s",
                         self.name, plugin_id, instance_id)
            plugin = self.config.add_source(
                plugin_id=plugin_id, instance_id=instance_id, priority=priority)

            # Configerus plugins all work differently so we take a different
            # approach per plugin
            if plugin_id == PLUGIN_ID_SOURCE_PATH:
                path = config_environment.get(
                    [instance_base, CONFIGERUS_PATH_KEY])
                plugin.set_path(path=path)
            elif plugin_id == PLUGIN_ID_SOURCE_DICT:
                data = config_environment.get(
                    [instance_base, CONFIGERUS_DICT_DATA_KEY])
                plugin.set_data(data=data)
            elif plugin_id == PLUGIN_ID_SOURCE_ENV_SPECIFIC:
                source_base = config_environment.get(
                    [instance_base, CONFIGERUS_ENV_SPECIFIC_BASE_KEY])
                plugin.set_base(base=source_base)
            elif plugin_id == PLUGIN_ID_SOURCE_ENV_JSON:
                source_env = config_environment.get(
                    [instance_base, CONFIGERUS_ENV_JSON_ENV_KEY])
                plugin.set_env(env=source_env)
            # this should probably be a configerus standard
            elif hasattr(plugin, 'set_data'):
                data = config_environment.get([instance_base, 'data'])
                plugin.set_data(data=data)
            else:
                logger.warning("had no way of configuring new Configerus source plugin %s",
                               plugin_id)

    def bootstrap(self, entrypoints: List[str]):
        """Bootstrap some METTA distributions.

        METTA bootstrapping is an attempt to allow an easy in to including contrib
        functionality without having to do a lot of Python imports.

        BootStrapping is a setuptools enabled process, where any python package can
        declare a bootstraper, and this function will run that bootstrapper on
        request.
        The BootStrap entry_points are expected to receive a config object on which
        they can operate to add any specific or global functionality.

        BootStraps are typically used for two behaviours:

        1. just import files which run configerus or metta decorators to register
            plugins
        2. add source/formatter/validator plugins to the passed config object.

        Parameters:
        -----------
        bootstrap (List[str]) : a list of string bootstrapper entry_points for the
            ucct.bootstrap entry_points (part of setuptools.)
            Each value needs to refer to a valid entrypoint which will be executed
            with the config object as an argument.

        Raises:
        -------
        Raises a KeyError in cases of a bootstrap ID that cannot be found.

        Bootstrappers themselves may raise an exception.

        """
        for entrypoint in entrypoints:
            if entrypoint not in self.bootstrapped:
                logger.debug("Running metta bootstrap entrypoint: %s ", entrypoint)
                for metta_ep in metadata.entry_points()[METTA_BOOTSTRAP_ENTRYPOINT]:
                    if metta_ep.name == entrypoint:
                        plugin = metta_ep.load()
                        plugin(self)
                        self.bootstrapped.append(entrypoint)
                        break
                else:
                    raise KeyError(f"Bootstrap not found {METTA_BOOTSTRAP_ENTRYPOINT}:{entrypoint}")
            else:
                logger.debug("Skipping boostrap %s, as it has already been imported", entrypoint)

    def plugin_priority(self, delta: int = 0):
        """Return a default plugin priority with a delta."""
        return self.default_priority + delta

    # Generic Plugin construction

    def add_fixtures_from_config(
            self, label: str = METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
            base: Union[str, List[Any]] = LOADED_KEY_ROOT, plugin_type: str = None,
            validator: str = '', exception_if_missing: bool = False,
            arguments: Dict[str, Any] = None) -> Fixtures:
        """Create plugin fixtures from some config.

        This method will interpret some config values as being usable to build a collection
        of plugins from.  The plugins will be built, wrapped as fixtures and added to the
        Environment.  The plugins will then be returned as a Fixtures collection.

        Parameters:
        -----------
        config (Config) : Used to load and get the plugin configuration

        label (str) : config label to load to pull plugin configuration. That
            label is loaded and config is pulled to produce a list of plugins

        base (str|List) : config key to get a Dict of plugins configurations.  This
            should point to a dict of plugin configurations.
            A list of strings is valid as configerus.loaded.get() can take that as
            an argument.
            We call this base instead of key as we will be searching for sub-paths
            to pull individual elements

        type (str) : plugin type to create, pulled from the config/dict if
            omitted.  If Type is provided by neither argument nor source then
            you're gonna have a bad time.

        validator (str) : optionally use a configerus validator on the instance
            config/dict before a plugin is created.

        arguments (Dict[str, Any]) : A Dict of named arguments to pass to the
            plugin constructor.  Constructors should be able to work without
            requiring the arguments, but these tend to be pivotal for them.

        Returns:
        --------
        Fixtures of your type

        Raises:
        -------
        If you ask for a plugin which has not been registered, then you're going to
        get a NotImplementedError exception.
        To make sure that your desired plugin is registered, make sure to import
        the module that contains the factory method with a decorator.

        """
        if arguments is None:
            arguments = {}

        fixtures = Fixtures()
        """ plugin set which will be used to create new plugins """

        try:
            plugin_config = self.config.load(label)
            plugin_list = plugin_config.get(base)
        except KeyError as err:
            if exception_if_missing:
                raise KeyError('Could not load any config for plugin generation') from err
            return fixtures

        for instance_id in plugin_list.keys():
            # This fixture gets effectively added to 2 different Fixtures object.
            # 1. we manually add it to our Fixtures object for this function call
            # 2. the add_fixture_from_config() adds it to the fixture for this
            #    environment object.
            fixture = self.add_fixture_from_config(
                label=label, base=[base, instance_id], plugin_type=plugin_type,
                instance_id=instance_id, validator=validator, arguments=arguments)
            fixtures.add(fixture)

        return fixtures

    def add_fixture_from_config(self, label: str, base: Union[str, List[Any]] = LOADED_KEY_ROOT,
                                plugin_type: str = None, instance_id: str = '', priority: int = -1,
                                validator: str = '', arguments: Dict[str, Any] = None) -> Fixture:
        """Create and add a new plugin fixture from some config.

        This method will interpret some config values as being usable to build a single
        plugin.  The plugin will be built, wrapped as a Fixture, and added to the environment.
        The plugin Fixture is returned.

        @see add_fixture_from_loadedconfig

        @note we could validate here instead of passing on the validator, but we
            pass it on to keep parity across a number of cases, and to keep as
            much complex logic in one place.

        Parameters:
        -----------
        config (Config) : Used to load and get the plugin configuration

        type (str) : plugin type to create, pulled from the config/dict if
            omitted. An exception will be thrown if Type is found from neither
            this argument nor the passed config.

        label (str) : config label to load to pull plugin configuration. That
            label is loaded and config is pulled to produce a list of plugins.

        base (str|List) : config key used as a .get() base for all gets.  With this
            you can instruct to pull config from a section of loaded config.
            A list of strings is valid because configerus.loaded.get() can take that
            as an argument. We will be using the list syntax anyway.
            We call this base instead of key as we will be searching for sub-paths
            to pull individual elements.

        instance_id (str) : optionally pass an instance_id for the item.

        validator (str) : optionally use a configerus validator on the entire .get()
            for the instance config.

        Returns:
        --------
        A Fixture object that wraps the created plugin.

        The Fixtures has already been added to the environment, but is returned
        so that the consumer can act on it separately without haveing to
        search for it.

        Raises:
        -------
        If you ask for a plugin which has not been registered, then you're going
        to get a NotImplementedError exception. To make sure that your desired
        plugin is registered, make sure to import the module that contains the
        factory method with a decorator.

        A ValueError is thrown if the plugin cannot be created due to missing
        plugin configuration/argument values.  This means that we could not
        determine how to create the plugin.

        A configerus.validate.ValidationError will be raised if a validation
        target was passed and validation failed.

        """
        if arguments is None:
            arguments = {}

        try:
            plugin_loaded = self.config.load(label)
            """ loaded configuration for the plugin """
        except KeyError as err:
            raise KeyError("Could not load plugin config source.") from err

        # If arguments were given then pass them on
        try:
            config_arguments = plugin_loaded.get([base, METTA_PLUGIN_CONFIG_KEY_ARGUMENTS])
            config_arguments.update(arguments)
            arguments = config_arguments
        except KeyError:
            if not arguments:
                # if no arguments were specified, then consider a special case of
                # 'from_config' which build arguments for a config based plugin
                # which will take a config label/base-key pair as config arguments
                #
                # There is a special case where if the passed from_config is not a
                # dict then the same config label/base received is used.
                try:
                    config_fromconfig = plugin_loaded.get(
                        [base, METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG])

                    if isinstance(config_fromconfig, dict):
                        logger.debug("Using from_config, and passing label/base as arguments")
                        arguments = config_fromconfig
                    else:
                        logger.debug(
                            "Using from_config, and passing current label/base as arguments")
                        arguments = {'label': label, 'base': base}
                except KeyError:
                    pass

        return self.add_fixture_from_loadedconfig(
            loaded=plugin_loaded, base=base, plugin_type=plugin_type, instance_id=instance_id,
            priority=priority, validator=validator, arguments=arguments)

    def add_fixture_from_dict(
            self, plugin_dict: Dict[str, Any], plugin_type: str = None, instance_id: str = '',
            validator: str = '', arguments: Dict[str, Any] = None) -> Fixture:
        """Create a single plugin from a Dict of information for it.

        Create a new plugin from a map/dict of settings for the needed parameters.

        @see add_fixture_from_loadedconfig

        Parameters:
        -----------
        config (Config) : configerus.Config object passed to each generated plugins.

        type (str) : plugin type to create, pulled from the config/dict if
            omitted

        client_dict (Dict[str,Any]) : Dict from which all needed information will
            be pulled.  Optionally additional config sources can be included as well
            as arguments which could be passed to the plugin.

            @see add_fixture_from_dict for more details.

        instance_id (str) : optionally pass an instance_id for the item.

        validator (str) : optionally use a configerus validator on the entire .get()
            for the instance config.

        Return:
        -------
        A Fixture object with the new plugin added

        The Fixtures has already been added to the environment, but is returned
        so that the consumer can act on it separately without haveing to
        search for it.

        """
        # Create a mock configerus.loaded.Loaded object, not attached to anything
        # and use it for config retrieval.  This gives us formatting, validation
        # etc.
        mock_config_loaded = Loaded(
            data=plugin_dict,
            parent=self.config,
            instance_id='mock-plugin-construct')
        """ Mock configerus loaded object for config retrieval """
        base = LOADED_KEY_ROOT
        """ to keep this function similar to add_fixture_from_config we use an empty .get() base """

        if arguments is None:
            arguments = {}

        return self.add_fixture_from_loadedconfig(
            loaded=mock_config_loaded, base=base, plugin_type=plugin_type, instance_id=instance_id,
            validator=validator, arguments=arguments)

    # This is where we centralize all logic around creating fixtures, so it is complex
    # pylint: disable=too-many-branches, too-many-locals
    def add_fixture_from_loadedconfig(
            self, loaded: Loaded, base: Union[str, List[Any]] = LOADED_KEY_ROOT,
            plugin_type: str = None, instance_id: str = '', priority: int = -1,
            validator: str = '', arguments: Dict[str, Any] = None) -> Fixture:
        """Create a plugin from a Configerus loaded config object.

        This method will interpret some config values as being usable to build plugin.
        This function starts with a loaded config object because we can leverage
        that from more starting points.

        Using a configerus config object allows us to leverage advanced configerus
        features such as tree searching, formatting and validation.

        What is looked for:

        1. valdiators if we need to validate the entire label/key before using it
        2. type if we did not receive a type
        3. plugin_id : which will tell us what plugin to load
        4. optional instance_id if none was passed
        5. config if you want config added - ONLY if fixtures is None
           (plugins in Fixtures cannot override config objects)
        6. arguments that will be executed on an argument() method if the
            plugin has it.

        Parameters:
        -----------
        config (Config) : Used to load and get the plugin configuration

        plugin_type (str) : plugin type to create, pulled from the config/dict if
            omitted. An exception will be thrown if type is found from neither
            this argument nor the passed config.

        label (str) : config label to load to pull plugin configuration. That
            label is loaded and config is pulled to produce a list of plugins.

        base (str|List) : config key used as a .get() base for all gets.  With this
            you can instruct to pull config from a section of loaded config.
            A list of strings is valid because configerus.loaded.get() can take that
            as an argument. We will be using the list syntax anyway.
            We call this base instead of key as we will be searching for sub-paths
            to pull individual elements.

        instance_id (str) : optionally pass an instance_id for the item.

        validator (str) : optionally use a configerus validator on the entire .get()
            for the instance config.

        Returns:
        --------
        A Fixture object with the new plugin added

        The Fixtures has already been added to the environment, but is returned
        so that the consumer can act on it separately without haveing to
        search for it.

        Raises:
        -------
        If you ask for a plugin which has not been registered, then you're going
        to get a NotImplementedError exception. To make sure that your desired
        plugin is registered, make sure to import the module that contains the
        factory method with a decorator.

        A ValueError is thrown if the plugin cannot be created due to missing
        plugin configuration/argument values.  This means that we could not
        determine how to create the plugin.

        A configerus.validate.ValidationError will be raised if a validation
        target was passed and validation failed.

        """
        # Retrieve all of the plugin config, to test that it exists
        # it might be expensive to retrieve all this but we do it to catch early
        # faults in config.
        plugin_base = loaded.get(base)
        """ configuration base object for the plugin """
        if plugin_base is None:
            raise ValueError("Cannot build plugin as provided config was empty.")

        # get any validators from config, defaulting to just the jsonschema
        # validator for a fixture
        validators = loaded.get(
            [base, METTA_PLUGIN_CONFIG_KEY_VALIDATORS],
            default=[{PLUGIN_ID_VALIDATE_JSONSCHEMA: METTA_FIXTURE_VALIDATION_JSONSCHEMA}])

        if validator:
            # if a validator arg was passed in, then add it
            validators.append(validator)
        if len(validators):
            # Run configerus validation on the config base once per validator
            try:
                for val in validators:
                    loaded.validate(plugin_base, validate_target=val)
            except ValidationError as err:
                raise err

        if not plugin_type:
            try:
                plugin_type = str(loaded.get([base, METTA_PLUGIN_CONFIG_KEY_PLUGINTYPE]))
            except KeyError as err:
                full_config = loaded.get(base)
                raise ValueError("Could not find a plugin type when trying to create a"
                                 f"plugin : {full_config}") from err

        try:
            plugin_id = str(loaded.get([base, METTA_PLUGIN_CONFIG_KEY_PLUGINID]))
        except KeyError as err:
            full_config = loaded.get(base)
            raise ValueError(
                "Could not find a plugin_id when trying to create a "
                f"'{type}' plugin from config: {full_config}") from err

        # if no instance_id was passed, try to load one or just make one up
        if not instance_id:
            instance_id = str(loaded.get([base, METTA_PLUGIN_CONFIG_KEY_INSTANCEID], default=''))
            if not instance_id:
                instance_rand = ''.join(random.choice(string.ascii_lowercase) for i in range(10))
                instance_id = f"{plugin_type}-{plugin_id}-{instance_rand}"

        if priority < 0:
            priority = int(loaded.get([base, METTA_PLUGIN_CONFIG_KEY_PRIORITY],
                                      default=self.plugin_priority()))
            """ instance priority - this is actually a stupid way to get it """

        config_arguments = loaded.get(
            [base, METTA_PLUGIN_CONFIG_KEY_ARGUMENTS], default={})
        if len(config_arguments) > 0:
            if arguments is None:
                arguments = {}
            else:
                # if we have config arguments from two different sources, then add the config args
                # to a copy of the function parameter args.  We use a copy so that we make no
                # context mistakes by altering a passed Dict that may get used for more than
                # one plugin.
                arguments = arguments.copy()

            arguments.update(config_arguments)

        # Use the factory to make the .fixtures.Fixture
        fixture = self.add_fixture(
            plugin_type=plugin_type, plugin_id=plugin_id, instance_id=instance_id,
            priority=priority, arguments=arguments)

        return fixture

    def add_fixture(self, plugin_type: str, plugin_id: str, instance_id: str, priority: int,
                    arguments: Dict[str, Any] = None) -> Fixture:
        """Create a new plugin from parameters.

        Parameters:
        -----------
        config (Config) : configerus.Config object passed to each generated plugins.

        type (str) : plugin type to create.

        plugin_id (str) : METTA plugin id for the plugin type, to tell us what plugin
            factory to load.

            @see .plugin.Factory for more details on how plugins are loaded.

        instance_id (str) : string instance id that will be passed to the new plugin
            object.

        priority (int) : Integer priority 1-100 for comparative prioritization
            between other plugins.

        arguments (Dict[str, Any]) : Arguments which should be passed to the
            plugin constructor after environment and instance_id

        Return:
        -------
        A Fixture object with the new plugin added

        The Fixtures has already been added to the environment, but is returned
        so that the consumer can act on it separately without haveing to
        search for it.

        Raises:
        -------
        NotImplementedError if you asked for an unregistered plugin_id/type

        """
        if arguments is None:
            arguments = {}

        if not (isinstance(plugin_type, str) and isinstance(plugin_id, str) and
                isinstance(instance_id, str) and isinstance(priority, int)):
            raise ValueError(f"Bad value types passed for creating a fixture: {plugin_type}"
                             f":{plugin_id}:{instance_id} ({priority})")

        fac = Factory(plugin_type, plugin_id)
        plugin = fac.create(self, instance_id, **arguments)
        fixture = self.fixtures.new(
            plugin=plugin, plugin_type=plugin_type, plugin_id=plugin_id, instance_id=instance_id,
            priority=priority)
        return fixture



"""

METTA Cli package.

"""

import logging

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment

from .base import METTA_PLUGIN_TYPE_CLI
from .config import ConfigCliPlugin
from .environment import EnvironmentCliPlugin
from .fixtures import FixturesCliPlugin
from .provisioner import ProvisionerCliPlugin

logger = logging.getLogger('metta.cli')


METTA_PLUGIN_ID_CLI_CONFIG = 'config'
""" cli plugin_id for the config plugin """


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_PLUGIN_ID_CLI_CONFIG)
def metta_plugin_factory_cli_config(
        environment: Environment, instance_id: str = ''):
    """Create a config cli plugin."""
    return ConfigCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_ENVIRONMENT = 'environment'
""" cli plugin_id for the environment plugin """


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_PLUGIN_ID_CLI_ENVIRONMENT)
def metta_plugin_factory_cli_environment(
        environment: Environment, instance_id: str = ''):
    """Create an environment cli plugin."""
    return EnvironmentCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_FIXTURES = 'fixtures'
""" cli plugin_id for the fixtures plugin """


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_PLUGIN_ID_CLI_FIXTURES)
def metta_plugin_factory_cli_fixtures(
        environment: Environment, instance_id: str = ''):
    """Create a fixtures cli plugin."""
    return FixturesCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_PROVISIONER = 'provisioner'
""" cli plugin_id for the provisioner plugin """


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_PLUGIN_ID_CLI_PROVISIONER)
def metta_plugin_factory_provisioner_config(
        environment: Environment, instance_id: str = ''):
    """Create a provisioner cli plugin."""
    return ProvisionerCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint BootStrapping ------


# pylint: disable=unused-argument
def bootstrap(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.

    """



"""

Metta CLI plugin for launchpad

Provided CLI commands that allow interacting with launchpad as a provisioner,
examining launchpad config and more.

"""

import logging

import yaml

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .provisioner import LaunchpadProvisionerPlugin, METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID

logger = logging.getLogger('metta.cli.launchpad')

METTA_LAUNCHPAD_CLI_PLUGIN_ID = "metta_launchpad"
""" metta plugin_id for the launchpad cli plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class LaunchpadCliPlugin(CliBase):
    """Fire command/group generator for launchpad plugin commands."""

    def fire(self):
        """Return CLI command group."""
        if self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_PROVISIONER,
                plugin_id=METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID,
                exception_if_missing=False) is not None:

            return {
                'contrib': {
                    'launchpad': LaunchpadGroup(self.environment)
                }
            }

        return {}


class LaunchpadGroup():
    """Base Fire command group for launchpad cli commands."""

    def __init__(self, environment: Environment):
        """Create launchpad command list object."""
        self.environment = environment

    def _select_provisioner(self, instance_id: str = '') -> LaunchpadProvisionerPlugin:
        """Pick a matching provisioner."""
        if instance_id:
            return self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id='metta_launchpad',
                instance_id=instance_id)

        # Get the highest priority provisioner
        return self.environment.fixtures.get(
            plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id='metta_launchpad')

    def info(self, provisioner: str = '', deep: bool = False):
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)

        provisioner_info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                provisioner_info.update(fixture.plugin.info(True))

        return cli_output(provisioner_info)

    def hosts(self, provisioner: str = '', deep: bool = False):
        """List the hosts in the cluster/"""
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        config = client.describe_config()

        if deep:
            host_list = config['spec']['hosts']
        else:
            host_list = []
            for host in config['spec']['hosts']:
                list_host = {
                    'role': host['role']
                }
                if 'ssh' in host:
                    list_host.update({
                        'is_windows': False,
                        'address': host['ssh']['address']
                    })
                if 'winrm' in host:
                    list_host.update({
                        'is_windows': True,
                        'address': host['winrm']['address']
                    })

                host_list.append(list_host)

        return cli_output(host_list)

    def exec(self, cmd: str, provisioner: str = '', host: int = 0):
        """ Exec a command."""
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        cmds = cmd.split(' ')

        client.exec(host_index=host, cmds=cmds)

    def exec_interactive(self, cmd: str, provisioner: str = '', host: int = 0):
        """ Exec a command."""
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        cmds = cmd.split(' ')

        client.exec_interactive(host_index=host, cmds=cmds)

    def connect(self, provisioner: str = '', host: int = 0):
        """Exec a command."""
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        client.exec_interactive(host_index=host, cmds=[])

    def client_config(self, provisioner: str = ''):
        """Get the rendered config from the client."""
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        return cli_output(client.describe_config())

    def version(self, provisioner: str = ''):
        """Output a launchpad cli report."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.client.version()

    def describe(self, report: str, provisioner: str = ''):
        """Output a launchpad cli report."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.client.describe(report)

    # pylint: disable=redefined-builtin
    def fixtures(self, provisioner: str = '', type: str = '', plugin_id: str = '',
                 instance_id: str = ''):
        """List all outputs."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        if not hasattr(provisioner_plugin, 'get_fixtures'):
            raise ValueError('This provisioner does not keep fixtures.')

        fixture_list = [{
            'type': fixture.plugin_type,
            'plugin_id': fixture.plugin_id,
            'instance_id': fixture.instance_id,
            'priority': fixture.priority,
        } for fixture in provisioner_plugin.get_fixtures(type=type, plugin_id=plugin_id,
                                                         instance_id=instance_id).to_list()]

        return cli_output(fixture_list)

    def config_file(self, provisioner: str = ''):
        """Dump the config file."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin

        try:
            with open(provisioner_plugin.config_file) as config_file:
                config_contents = yaml.load(config_file)
        except FileNotFoundError as err:
            raise ValueError("No config file was found") from err

        return cli_output(config_contents)

    def write_config(self, provisioner: str = ''):
        """Write config to file."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        # access private method for introspection
        # pylint: disable=protected-access
        provisioner_plugin._write_launchpad_file()

    def client_bundle(self, provisioner: str = '',
                      user: str = 'admin', reload: bool = False):
        """Tell Launchpad to download the client bundle."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner)
        plugin = provisioner_plugin.plugin

        logger.info("Downloading client bungle for user: %s", user)
        try:
            # access private method for manual interaction
            # pylint: disable=protected-access
            bundle = plugin._mke_client_bundle(user=user, reload=reload)
        except Exception as err:
            raise Exception("Launchpad command failed.") from err

        return cli_output(bundle)

    def prepare(self, provisioner: str = ''):
        """Run provisioner prepare."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.prepare()

    def apply(self, provisioner: str = '', debug: (bool) = False):
        """Run provisioner apply."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.apply(debug=debug)

    def destroy(self, provisioner: str = '', quick: bool = False):
        """Run provisioner destroy."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.destroy(quick=quick)



"""

Metta contrib package for using Mirantis launchpad to provision clusters.

This allows using launchpad to install the Mirantis products onto a cluster
that had already been provisioned by other services.

This module registers the package metta plugins.

"""
from typing import Any

from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta_cli.base import METTA_PLUGIN_TYPE_CLI

from .launchpad import LaunchpadClient
from .provisioner import (LaunchpadProvisionerPlugin, METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID,
                          METTA_LAUNCHPAD_CONFIG_LABEL)
from .exec_client import (LaunchpadExecClientPlugin, METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID,)
from .cli import LaunchpadCliPlugin, METTA_LAUNCHPAD_CLI_PLUGIN_ID


@Factory(plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id=METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID)
def metta_plugin_factory_provisioner_launchpad(
        environment: Environment, instance_id: str = "", label: str = METTA_LAUNCHPAD_CONFIG_LABEL,
        base: Any = LOADED_KEY_ROOT):
    """Create a launchpad provisioner plugin."""
    return LaunchpadProvisionerPlugin(environment, instance_id, label, base)


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID)
def metta_terraform_factory_cliexec_client_launchpad(
        environment: Environment, instance_id: str = '', client: LaunchpadClient = None):
    """Create an launchpad exec client plugin."""
    return LaunchpadExecClientPlugin(environment, instance_id, client)


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_LAUNCHPAD_CLI_PLUGIN_ID)
def metta_terraform_factory_cli_launchpad(environment: Environment, instance_id: str = ''):
    """Create an launchpad cli plugin."""
    return LaunchpadCliPlugin(environment, instance_id)


# ----- METTA BOOTSTRAPPERS -----


# pylint: disable=unused-argument
def bootstrap(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.

    """



"""

Metta CLI : Fixture commands.

Cli plugin that allows examination of fixtures in environments.

"""

import logging

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment

from .base import CliBase, cli_output, METTA_PLUGIN_TYPE_CLI

logger = logging.getLogger('metta.cli.fixtures')


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class FixturesCliPlugin(CliBase):
    """Fire command/group generator for fixture commands."""

    def fire(self):
        """Return a dict of commands."""
        return {
            'fixture': FixturesGroup(self.environment)
        }


class FixturesGroup():
    """Base Fire command group for fixtures commands."""

    def __init__(self, environment: Environment):
        """Attach environment to object."""
        self._environment = environment

    def list(self, plugin_type: str = '', plugin_id: str = '', instance_id: str = ''):
        """Return List all fixture instance_ids."""
        fixture_list = []
        for fixture in self._environment.fixtures.filter(
            plugin_type=plugin_type,
            plugin_id=plugin_id,
            instance_id=instance_id):

            fixture_list.append(fixture.instance_id)

        return cli_output(fixture_list)

    # needs to be a method for registration in fire
    # pylint: disable=no-self-use
    def plugin_types(self):
        """List plugins that have been registered with the environment."""
        plugin_list = {}
        for plugin_type in Factory.registry:
            plugin_list[plugin_type] = []
            for plugin_id in Factory.registry[plugin_type]:
                plugin_list[plugin_type].append(plugin_id)

        return cli_output(plugin_list)

    # just lots of optional filtering arguments
    # pylint: disable=too-many-arguments
    def info(self, deep: bool = False, include_cli_plugins: bool = False):
        """Return Info for all fixtures."""
        fixture_info_list = []
        for fixture in self._environment.fixtures:

            if not include_cli_plugins and fixture.plugin_type is METTA_PLUGIN_TYPE_CLI:
                continue

            info = {
                'fixture': {
                    'plugin_type': fixture.plugin_type,
                    'plugin_id': fixture.plugin_id,
                    'instance_id': fixture.instance_id,
                    'priority': fixture.priority,
                }
            }

            if deep and hasattr(fixture.plugin, 'info'):
                plugin_info = fixture.plugin.info()
                if isinstance(plugin_info, dict):
                    info.update(plugin_info)

            fixture_info_list.append(info)

        return cli_output(fixture_info_list)



"""

Metta Docker.

metta contrib package for docker functionality, specifically for registering
a Docker client plugin.

"""

from typing import Any

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta.workload import METTA_PLUGIN_TYPE_WORKLOAD

from .client import DockerPyClientPlugin, METTA_PLUGIN_ID_DOCKER_CLIENT
from .run_workload import (DockerPyRunWorkloadPlugin, DOCKER_RUN_WORKLOAD_CONFIG_LABEL,
                           DOCKER_RUN_WORKLOAD_CONFIG_BASE)


# This is really what it takes to configure both the metta plugin and the docker client
# pylint: disable=too-many-arguments
@Factory(plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_PLUGIN_ID_DOCKER_CLIENT)
def metta_plugin_factory_client_dockerpy(
    environment: Environment, instance_id: str = '', host: str = '', cert_path: str = '',
        tls_verify: bool = True, compose_tls_version: str = 'TLSv1_2', version: str = 'auto'):
    """Create an metta client for dockerpy."""
    return DockerPyClientPlugin(environment, instance_id=instance_id,
                                host=host, cert_path=cert_path, tls_verify=tls_verify,
                                compose_tls_version=compose_tls_version, version=version)


METTA_PLUGIN_ID_DOCKER_RUN_WORKLOAD = 'metta_docker_run'
""" workload plugin_id for the docker run plugin."""


@Factory(plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id=METTA_PLUGIN_ID_DOCKER_RUN_WORKLOAD)
def metta_plugin_factory_workload_docker_run(environment: Environment, instance_id: str = '',
                                             label: str = DOCKER_RUN_WORKLOAD_CONFIG_LABEL,
                                             base: Any = DOCKER_RUN_WORKLOAD_CONFIG_BASE):
    """Create a docker run workload plugin."""
    return DockerPyRunWorkloadPlugin(environment, instance_id, label=label, base=base)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config
        added to.

    """



"""

METTA Core toolset.

In this module, the root package, are found a number of tool plugin constructors
that can be used to get instances of the plugins based on either passed in data
or configerus config.

The contstruction.py module does the heavy lifting of building plugins and
PluginInstances sets.

"""

import os
from typing import List
import logging
import importlib

from configerus import new_config as configerus_new_config
from configerus.config import Config
from configerus.loaded import LOADED_KEY_ROOT

from .globals import all_environments
from .discover import (discover_project_root, discover_sources_from_config,
                       discover_imports)
from .environment import Environment, METTA_PLUGIN_CONFIG_LABEL_ENVIRONMENTS

logger = logging.getLogger('metta')

""" METTA and Configerus bootstrapping """

FIXED_CONFIGERUS_BOOSTRAPS = [
    "get",
    "env",
    "jsonschema",
    "files"
]
""" configerus bootstraps that we will use on config objects. We use functionality
    in the core which heavily depends on this config, so it is necessary """
FIXED_METTA_BOOTSTRAPS = [
    "metta_common"
]
""" Metta bootstraps to apply to any created environemnt.  We use functionality
    in the core which heavily depends on these, so they are necessary """
DEFAULT_ADDITIONAL_METTA_BOOTSTRAPS = [
    "metta_ansible",
    "metta_docker",
    "metta_dummy",
    "metta_kubernetes",
    "metta_terraform"
]
""" default metta bootstrap calls to add to any created cluster """

DEFAULT_ENVIRONMENT_NAME = 'default'
""" If you don't provide a particular environment name """

CWD = os.path.realpath(os.getcwd())
""" CWD if needed for project discovery """

# ----- Environment interaction-----


def discover(path: str = CWD,
             additional_configerus_bootstraps: List[str] = None,
             additional_metta_bootstraps: List[str] = None):
    """Discover project config, environment and fixtures.

    If you want to just write some config and get defined config and
    environments, use this.

    This will:

    1. create a new config object
    2. look for 'metta.yml|json' files which define where your project root is
    3. add projects roots as config sources
    4. look for config that tells it to add more config sources
    5. look for config that tells it to import some python modules
    6. look for config that tells it to create some environments

    You should have:

    1. a metta.yml file in your project root (should not be empty)
        a. if you don't want to keep config in your project root then declare
           an additional config source for a config path
    2. an environments.yml file to declare environments and environment
       fixtures.


    Returns:
    --------
    Nothing.  This will create environments, which you can access by name.

    """
    if additional_configerus_bootstraps is None:
        additional_configerus_bootstraps = []
    if additional_metta_bootstraps is None:
        additional_metta_bootstraps = DEFAULT_ADDITIONAL_METTA_BOOTSTRAPS

    logger.info("Creating project from path %s", path)
    configerus_bootstraps = FIXED_CONFIGERUS_BOOSTRAPS + \
        additional_configerus_bootstraps
    config = configerus_new_config(bootstraps=configerus_bootstraps)

    # Look for project root, and root config (and look in that root config for
    # stuff to do)
    discover_project_config(config=config, path=path)
    # Create any environments defined in config

    metta_config = config.load('metta')

    if metta_config.has(['environments', 'from_config']):
        environment_config = metta_config.get(['environments', 'from_config'])
        label = environment_config['label'] if 'label' in environment_config else 'metta'
        base = environment_config['base'] if 'base' in environment_config else LOADED_KEY_ROOT
        new_environments_from_config(
            config=config,
            additional_metta_bootstraps=additional_metta_bootstraps,
            label=label,
            base=base)

    elif metta_config.has(['environments', 'from_builder']):
        environment_config = metta_config.get(['environments', 'from_builder'])
        module = environment_config['import']
        method = environment_config['method']
        new_environments_from_builder(
            config=config,
            additional_metta_bootstraps=additional_metta_bootstraps,
            module=module,
            method=method)

    elif metta_config.has('environments'):
        new_environments_from_config(
            config=config,
            additional_metta_bootstraps=additional_metta_bootstraps,
            label='metta',
            base='environments')


def discover_project_config(config: Config, path: str = CWD):
    """Discover project root and root config.

    We start of by scanning the path, and all of its parents for a metta.yml
    file.  We collect all of the paths as config sources, into a config object.

    We ask the config object if there are any additional config sources to add
    after adding all of the root config sources.

    We ask the config object for information about required python imports to
    allow local code to be imported.  This can be used to allow plugin
    registration.

    Parameters:
    -----------
    path (str) : we will be scanning path/parents looking for some configuration
      for meta.  With this argument you can suggest the starting poing for the
      scan.

    additional_configerus_bootstraps (List[str]) : run additional configerus
        bootstrap entry_points

    """
    logger.info("Discovering project config from path %s", path)

    # First see if we can find a root metta.yml file for file context
    # if we do, then we add config sources for such paths
    discover_project_root(config, path)
    # existing config may be able to tell us to add more config sources
    discover_sources_from_config(config)
    # import any module imports requested
    discover_imports(config)


def new_environment(
        name: str = DEFAULT_ENVIRONMENT_NAME,
        additional_metta_bootstraps: List[str] = None,
        additional_configerus_bootstraps: List[str] = None,
        load_more_from_label: str = '',
        load_more_from_base: str = LOADED_KEY_ROOT):
    """Make new environment object after making a new config.

    First create a config object, then use it to create an environment
    object.

    Parameters
    ----------
    additional_metta_bootstraps (List[str]) : run additiional metta bootstraps
        on the config object.  Defaults to the metta bootstrap.

    additional_configerus_bootstraps (List[str]) : run additional configerus
        bootstrap entry_points

    # Optionally tell the environment to get more information about itself from
    # a config label/key root

    load_more_from_label (str) : Config label to load to get more info about
        the environment

    load_more_from_base (str) : Config key to .get() as a base for more info
        about the environment

    Returns:
    --------
    An initialized Environment object with a new configerus Config object

    """
    if additional_metta_bootstraps is None:
        additional_metta_bootstraps = DEFAULT_ADDITIONAL_METTA_BOOTSTRAPS
    if additional_configerus_bootstraps is None:
        additional_configerus_bootstraps = []

    logger.info("Creating single environment from config")
    configerus_bootstraps = FIXED_CONFIGERUS_BOOSTRAPS + \
        additional_configerus_bootstraps
    config = configerus_new_config(bootstraps=configerus_bootstraps)

    metta_bootstraps = FIXED_METTA_BOOTSTRAPS + additional_metta_bootstraps
    environment = Environment(
        name=name,
        config=config,
        bootstraps=metta_bootstraps,
        config_label=load_more_from_label,
        config_base=load_more_from_base)
    return add_environment(name, environment)


def new_environments_from_config(
        config: Config,
        additional_metta_bootstraps: List[str] = None,
        label: str = METTA_PLUGIN_CONFIG_LABEL_ENVIRONMENTS,
        base: str = LOADED_KEY_ROOT):
    """Create new environments from config.

    Ask the passed config object for information about environments.  From that
    information we create an environment objects.

    Parameters:
    -----------
    config (Config) : config source used to discover environments to create, and
        passed to environments objects for construction

    additional_metta_bootstraps (List[str]) : run additiional metta bootstraps
        on the config object.  Defaults to the metta bootstrap.

    additional_configerus_bootstraps (List[str]) : run additional configerus
        bootstrap entry_points

    """
    if additional_metta_bootstraps is None:
        additional_metta_bootstraps = DEFAULT_ADDITIONAL_METTA_BOOTSTRAPS.copy()

    logger.info("Discovering environments from config: %s:%s", label, base)

    # Build any environment describe in config (in all of the metta.yml files)
    environments_config = config.load(label)
    if environments_config is not None:
        environments_dict = environments_config.get(base)
        if environments_dict is not None:
            metta_bootstraps = FIXED_METTA_BOOTSTRAPS + additional_metta_bootstraps
            for name in environments_dict.keys():
                environment_base = [base, name]
                logger.debug("creating new environment from config: %s:%s", label, environment_base)
                environment = Environment(
                    name=name,
                    config=config,
                    bootstraps=metta_bootstraps,
                    config_label=label,
                    config_base=environment_base)
                add_environment(name, environment)


def new_environment_from_config(
        config: Config,
        name: str = DEFAULT_ENVIRONMENT_NAME,
        additional_metta_bootstraps: List[str] = None,
        load_more_from_label: str = '',
        load_more_from_base: str = LOADED_KEY_ROOT):
    """Make a new environment from an existing configerus.Config object.

    Use a passed configerus Config object to create an environment object,
    register the env and then return it.

    The config object is bootstrapped before we create the environment.

    Parameters
    ----------
    config (Config) : passed to the environment for construction

    additional_metta_bootstraps (List[str]) : run additiional metta bootstraps
        on the config object.  Defaults to the metta bootstrap.

    additional_configerus_bootstraps (List[str]) : run additional configerus
        bootstrap entry_points

    Returns:
    --------
    An intialized Environment object from the config

    """
    if additional_metta_bootstraps is None:
        additional_metta_bootstraps = DEFAULT_ADDITIONAL_METTA_BOOTSTRAPS.copy()

    bootstraps = FIXED_METTA_BOOTSTRAPS + additional_metta_bootstraps
    environment = Environment(
        name=name,
        config=config,
        bootstraps=bootstraps,
        config_label=load_more_from_label,
        config_base=load_more_from_base)
    return add_environment(name, environment)


def new_environments_from_builder(
        config: Config, additional_metta_bootstraps: List[str], module: str, method: str):
    """Create environments using an external environment builder option.

    This method will import and call a method to build environments dynamically.

    This is usefull in cases where building environments from flat config is not
    possible due to complexity, or if it would be overly redundant.

    This method receives a python package/module name which it imports.  Then the
    method of that imported package is executed and given the Config parameter.
    The method is expected to operate independently, using metta functions to
    register any environments it creates.

    A common approach for this is to use a builder to build dynamic sets of
    configuration from simpler sources, and then use the other new_environment
    methods to process that config.

    Parameters:
    -----------
    config (Config) : base config object which can be used as a base for any new
        environment objects, but can also be used as the source of configuration
        for helping the builder to decide what environments to make.

    additional_metta_bootstraps (List[str]) : metta bootstraps to include on any
        built environment.  This doesn't really make sense, but it is a part
        of the building paradigm so it made sense to included it here.

    module (string) : Python module containing the builder.  This will be
        imported.

    method (string) : Python method which will create the environments.

    """
    try:
        module = importlib.import_module(module)
    except Exception as err:
        raise ValueError(f"Env builder could not load suggested python package:{err}") from err

    try:
        method = getattr(module, method)
    except Exception as err:
        raise ValueError(f"Env builder could not execute package method: {err}") from err

    method(config, additional_metta_bootstraps)


def environment_names() -> List[str]:
    """Return a list of all of the created environments."""
    return list(all_environments)


def has_environment(name: str) -> bool:
    """Check if an environment already exist with a passed name."""
    return name in all_environments.keys()


def get_environment(name: str = '') -> Environment:
    """Return an environment that has already been created."""
    try:
        if name == '':
            name = list(all_environments.keys())[0]
        return all_environments[name]
    except KeyError as err:
        raise KeyError(f"Requested environment has not yet been created: {name}") from err


def add_environment(name: str, environment: Environment):
    """Add an environment to the global list."""
    if name in all_environments:
        logger.warning("Existing environment '%s' is being overwritten", name)
    all_environments[name] = environment
    return environment



"""

Metta kubernetes CLI plugin.

Provides functionality to manually interact with the various kubernetes plugins.

"""
import logging
from typing import Dict, Any

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta.workload import METTA_PLUGIN_TYPE_WORKLOAD
from mirantis.testing.metta_cli.base import CliBase, cli_output

logger = logging.getLogger('metta.cli.kubernetes')


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class KubernetesCliPlugin(CliBase):
    """Fire command/group generator for various kubernetes plugin commands."""

    def fire(self) -> Dict[str, Any]:
        """Return command groups for Kubernetes plugins."""
        if self.environment.fixtures.get(plugin_type=METTA_PLUGIN_TYPE_CLIENT,
                                         plugin_id='metta_kubernetes',
                                         exception_if_missing=False) is not None:

            return {
                'contrib': {
                    'kubernetes': KubernetesGroup(self.environment)
                }
            }

        return {}


class KubernetesGroup():
    """Base Fire command group for terraform client cli commands."""

    def __init__(self, environment: Environment):
        """Add additional command groups for plugins and inject environment."""
        self._environment = environment

        if self._environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id='metta_kubernetes_yaml',
                exception_if_missing=False) is not None:
            self.yaml = KubernetesYamlWorkloadGroup(self._environment)
        if self._environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id='metta_kubernetes_helm',
                exception_if_missing=False) is not None:
            self.helm = KubernetesHelmWorkloadGroup(self._environment)

    def _select_client(self, instance_id: str = '') -> Fixtures:
        """Pick a matching client."""
        if instance_id:
            return self._environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id='metta_kubernetes',
                instance_id=instance_id)

        # Get the highest priority workload
        return self._environment.fixtures.get(
            plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id='metta_kubernetes')

    def info(self, workload: str = '', deep: bool = False):
        """Get info about a client plugin."""
        fixture = self._select_client(instance_id=workload)

        collect_info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                collect_info.update(fixture.plugin.info(True))

        return cli_output(collect_info)

    def readyz(self, workload: str = ''):
        """Get kubernetes readiness info from the plugin."""
        plugin = self._select_client(instance_id=workload).plugin

        try:
            return cli_output(plugin.readyz())

        except Exception as err:
            raise RuntimeError('Kubernetes is not ready') from err

    def livez(self, workload: str = ''):
        """Get kubernetes livez info from the plugin."""
        plugin = self._select_client(instance_id=workload).plugin

        try:
            return cli_output(plugin.livez())

        except Exception as err:
            raise RuntimeError("Kubernetes is not ready") from err

    def connect_service_proxy(self, namespace: str,
                              service: str, workload: str = ''):
        """Create a service proxy."""
        plugin = self._select_client(instance_id=workload).plugin

        try:
            # Thank you kubernetes for the naming pattern
            # pylint: disable=invalid-name
            CoreV1Api = plugin.get_api('CoreV1Api')
            sc = CoreV1Api.connect_post_namespaced_service_proxy(namespace=namespace, name=service)

            return cli_output(sc)

        except Exception as err:
            raise RuntimeError("Exception trying to open the service proxy") from err


class KubernetesYamlWorkloadGroup():
    """Base Fire command group for terraform yaml workload plugin cli commands."""

    def __init__(self, environment: Environment):
        """Inject environment into command group object."""
        self._environment = environment

    def _select_fixture(self, instance_id: str = '') -> Fixtures:
        """Pick a matching workload fixture."""
        if instance_id:
            return self._environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id='metta_kubernetes_yaml',
                instance_id=instance_id)

        # Get the highest priority workload
        return self._environment.fixtures.get(
            plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id='metta_kubernetes_yaml')

    def info(self, workload: str = '', deep: bool = False):
        """Get info about a yaml workload plugin."""
        fixture = self._select_fixture(instance_id=workload)

        collect_info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                collect_info.update(fixture.plugin.info(True))

        return cli_output(collect_info)

    def apply(self, workload: str = ''):
        """Run workload apply."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        instance = workload_plugin.create_instance(self._environment.fixtures)

        objects = instance.apply()

        return cli_output(objects)

    def destroy(self, workload: str = ''):
        """Run workload destroy."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        instance = workload_plugin.create_instance(self._environment.fixtures)

        destroy = instance.destroy()

        return cli_output(destroy)


class KubernetesHelmWorkloadGroup():
    """Base Fire command group for terraform helm workload cli commands."""

    def __init__(self, environment: Environment):
        """Inject environment into command group."""
        self._environment = environment

    def _select_fixture(self, instance_id: str = ''):
        """Pick a matching workload fixture."""
        if instance_id:
            return self._environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id='metta_kubernetes_helm',
                instance_id=instance_id)

        # Get the highest priority workload
        return self._environment.fixtures.get(
            plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id='metta_kubernetes_helm')

    def info(self, workload: str = '', deep: bool = False) -> str:
        """Get info about a helm workload plugin."""
        fixture = self._select_fixture(instance_id=workload)

        collect_info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                collect_info.update(fixture.plugin.info(True))

        return cli_output(collect_info)

    def apply(self, workload: str = '', wait: bool = True,
              debug: bool = False) -> str:
        """Run helm workload apply."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        instance = workload_plugin.create_instance(self._environment.fixtures)

        objects = instance.apply(wait=wait, debug=debug)

        return cli_output(objects)

    def destroy(self, workload: str = '', debug: bool = False) -> str:
        """Run helm workload destroy."""
        workload_plugin = self._select_fixture(instance_id=workload).plugin
        instance = workload_plugin.create_instance(self._environment.fixtures)

        destroy = instance.destroy(debug=debug)

        return cli_output(destroy)



"""

CLI Plugins base.

@TODO move to the metta_cli package

Define a base class and some constants for the CLI plugins.  The CLI plugins
provide functionality for the metta cli tool, which is provided by the
metta_cli package.  This is here because the plugin types are defined as an
enum in this package, and so the base class is here too.

"""
import logging
import json
from typing import Any

from mirantis.testing.metta.environment import Environment

logger = logging.getLogger('metta.cli.base')

METTA_PLUGIN_TYPE_CLI = "cli"
""" Metta plugin_type identifier for CLI plugins """


# pylint: disable=too-few-public-methods
class CliBase:
    """Base class/interface for cli plugins.

    Mainly here as an interface definition so that you know what you have to
    add to get it to work with the metta_cli system.

    """

    def __init__(self, environment: Environment, instance_id: str):
        """Inject Environment and instance_id into plugin."""
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

    def fire(self):
        """Execute the fire cli hook."""
        raise NotImplementedError("this functionality not avialble.")


def cli_output(structure: Any) -> str:
    """Format data for output from cli commands.

    Parameters:
    -----------
    structure (Any) : data to be formatted for cli return

    Returns:
    --------
    String conversion, as json

    """
    return json.dumps(structure, indent=2, default=_serialize_last_resort)


def _serialize_last_resort(target: Any) -> str:
    """Serialzie anything.

    Returns:
    --------
    This target as a string.  Serialized for output only.

    """
    return f"{target}"



"""

Common configuration used for mirantis metta functionality integration.

The mirantis_metta module is based around custom functionality which we use to
extend metta, but which may not be desired by all users. This common core
allows integration of common config used for that functionalty.

"""

import logging
import pkg_resources

from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH as CONFIGERUS_SOURCE_PATH
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT as CONFIGERUS_SOURCE_DICT

from mirantis.testing.metta.environment import Environment

logger = logging.getLogger("metta_mirantis.common")

METTA_MIRANTIS_COMMON_NAME = "metta-mirantis"
""" used for some path building for config sources """
METTA_MIRANTIS_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS = 35
""" Config source priority for added common config """
METTA_MIRANTIS_PATH = pkg_resources.resource_filename(
    "mirantis.testing.metta_mirantis", "")
""" Path to the metta mirantis project root """
METTA_MIRANTIS_CONFIG_PATH = pkg_resources.resource_filename(
    'mirantis.testing.metta_mirantis', "config")
""" Path to the metta mirantis config preset configurations """


def add_common_config(environment: Environment):
    """Add some common configuration sources.

    The following could be added:

    1. a path config key for a path to this module.  This is used to make it
       easier to build paths to things like the terraform plans in this package
    2. add this module's ./config path as a config source.

    Parameters:
    -----------
    environment (Environment) : METTA environment to be modified

    """
    # If we haven't already done it, add the metta_mirantis/config path as a
    # config source.  This allows us to provide config defaults in this module
    metta_mirantis_config_instance_id = f"{METTA_MIRANTIS_COMMON_NAME}-config"
    if not environment.config.has_source(metta_mirantis_config_instance_id):
        source = environment.config.add_source(
            CONFIGERUS_SOURCE_PATH,
            metta_mirantis_config_instance_id,
            priority=METTA_MIRANTIS_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS)
        source.set_path(METTA_MIRANTIS_CONFIG_PATH)

    # Add a "{path:metta_mirantis}" config value so that paths to this module
    # can be built.  We use this to allow file resources from this module to be
    # used as a part of config.  For example we build paths to the terraform
    # plans in this module.
    metta_mirantis_paths_instance_id = f"{METTA_MIRANTIS_COMMON_NAME}-paths"
    if not environment.config.has_source(metta_mirantis_paths_instance_id):
        source = environment.config.add_source(
            CONFIGERUS_SOURCE_DICT, metta_mirantis_paths_instance_id,
            priority=METTA_MIRANTIS_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS)
        source.set_data(
            {
                environment.config.paths_label(): {
                    METTA_MIRANTIS_COMMON_NAME: METTA_MIRANTIS_PATH
                }
            })



"""

Metta client plugin for kubernetes.

This client makes a strong effort to behave as an extension of the community
kubernetes python library, combining this with the metta config system, and
allowing creation by any provisioning source.


"""
import logging
import re
import time
from typing import Dict, Any

import kubernetes

logger = logging.getLogger('metta.contrib.kubernetes.client.kubeapi')

METTA_PLUGIN_ID_KUBERNETES_CLIENT = 'metta_kubernetes'
""" client plugin_id for the metta kubernetes client plugin """

KUBEAPI_CLIENT_INTERPRET_Z_REGEX = re.compile(
    r'^\[(?P<symbol>[+-])\](?P<name>\S+)\s{1}(?P<ok>\w+)$')


class KubernetesApiClientPlugin:
    """Metta Client plugin for Kubernetes.

    Construction:
    -------------
    As an metta Plugin, the class constructor is used to initialize the
    instance as a plugin.  This means that we lose the constructor, but we gain
    access to an metta.config.Config object.

    We use .args() to configure the instance and create the client, as per the
    client plugin standard.

    To use:
    -------

    2. Ask metta.client.make_client for an instance of the plugin
        a. it is more expected that a provisioner will provide the client
           directly
        b. you can ask for the plugin directly from metta.plugin.get_plugin()
        c. you can use metta.new_clients_from_config() if you have some
           configuration with the needed args, and it will build clients for
           you
    3. Configure using .args() adding a kubeconfig file with activated context
       if that wasn't done for you when you received the client
    4. Ask the client for a kubernetes API version client
       (such as get_api('CoreV1Api') => CoreV1Api)
    5. Use the kubernetes API client as normal

    ```
    import mirantis.testing.metta as metta
    import mirantis.testing.metta.client as metta_client
    import mirantis.testing.metta_kubernetes as metta_kubernetes

    config = metta.new_config()
    client = metta_client.make_client(
                metta_kubernetes.METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                config, 'my-k8-instance')

    client.args('path/to/k8file')

    core_v1 = client.get_api('CoreV1Api')
    ns = core_v1.read_namespace(name="kube-system")
    print(f"NS: {ns}")
    ```

    Why use this:
    -------------
    It is cumbersome to use this if you already have access to the kubeconfig,
    but it fits into the plugin system, and therefore auto-loading and
    auto-configuring is possible with this implementation.
    Effectively, it is not easier to use this plu-gin over the native K8 client
    unless you consider that a provisioner can provide you this plugin already
    configured with its own information.

    This plugin is particularly usefull as it can be used for the helm and
    deployment workload plugins directly, which know how to use it to apply
    workloads to a kubernetes cluster.

    """

    def __init__(self, environment, instance_id, kube_config_file: str = ''):
        """Run the super constructor but also set class properties.

        This implements the args part of the client interface.

        Here we expect to receive a path to a KUBECONFIG file with a context set
        and we create a Kubernetes client for use.  After that this can provide
        Core api clients as per the kubernetes SDK

        Parameters:
        -----------
        config_file (str): String path to the kubernetes config file to use

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        logger.debug("Creating Kuberentes client from config file")
        self.api_client = kubernetes.config.new_client_from_config(
            config_file=kube_config_file)
        """ Kubernetes api client """

        self.config_file = kube_config_file
        """ Kube config file, in case you need to steal it. """

    def info(self) -> Dict[str, Any]:
        """Return dict data about this plugin for introspection."""
        return {
            'kubernetes': {
                'config_file': self.config_file
            }
        }

    def get_api(self, api: str):
        """Get an kubernetes API."""
        if hasattr(kubernetes.client, api):
            return getattr(kubernetes.client, api)(self.api_client)

        raise KeyError(f"Unknown API requested: {api}")

    def utils_create_from_yaml(self, yaml_file: str, **kwargs):
        """Run a kube apply from a yaml file."""
        return kubernetes.utils.create_from_yaml(
            k8s_client=self.api_client, yaml_file=yaml_file, **kwargs)

    def utils_create_from_dict(self, data: Dict, **kwargs):
        """Run a kube apply from dict of K8S yaml."""
        return kubernetes.utils.create_from_dict(
            k8s_client=self.api_client, data=data, **kwargs)

    def ready_wait(self, timeout: int = 30, period: int = 1):
        """Wait until kubernetes is ready before returning."""
        err = None
        while timeout > 0:
            try:
                ready = self.readyz()
                return ready
            except kubernetes.client.rest.ApiException as this_err:
                err = this_err
                time.sleep(period)
                timeout = timeout - period
                continue

        raise RuntimeError('Timed out waiting for kubernetes to become ready') from err

    def readyz(self):
        """Check the general readyz endpoint.

        Returns:
        --------
        Dict of service:status_dict with values symbol (+) and string ok value

        Raises:
        -------
        Will raise an exception is kubernetes isn't avaialble

        """
        return self._interpret_z_response('/readyz')

    def livez(self):
        """Check the general livez endpoint.

        Returns:
        --------
        Dict of service:status_dict with values symbol (+) and string ok value

        Raises:
        -------
        Will raise an exception is kubernetes isn't avaialble

        """
        return self._interpret_z_response('/livez')

    def _interpret_z_response(self, endpoint: str, method: str = 'GET',
                              params: Dict[str, str] = None) -> Dict[str, Dict[str, str]]:
        """Interpret that readyz/livez response format into a dict.

        because both livez and readyz return similar a machine un-friendly response, we use this
        method to retrieve and interpret the result.

        """
        # this will produce an exception if K8s is not ready
        response = self.api_client.call_api(
            method=method,
            resource_path=endpoint,
            query_params=params if params is not None else {},
            _preload_content=False)[0]

        interpreted = {}
        for line in response.read().decode("utf-8") .split('\n'):
            match = KUBEAPI_CLIENT_INTERPRET_Z_REGEX.fullmatch(line)
            if match:
                interpreted[match.group('name')] = {
                    'symbol': match.group('symbol'),
                    'ok': match.group('ok')
                }

        return interpreted

    # this allows consumers to get a watch without importing kubernetes
    # pylint: disable=no-self-use
    def watch(self):
        """Get a kubernetes watch instance."""
        return kubernetes.watch.Watch()



"""

Metta testkit functionality.

this module registers the testkit metta plugin factories to make them avaialble for use.

"""
from typing import Any

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER
from mirantis.testing.metta_cli.base import METTA_PLUGIN_TYPE_CLI

from .provisioner import (TestkitProvisionerPlugin, METTA_PLUGIN_ID_TESTKIT_PROVISIONER,
                          TESTKIT_PROVISIONER_CONFIG_LABEL, TESTKIT_PROVISIONER_CONFIG_BASE)
from .cli import TestkitCliPlugin, METTA_PLUGIN_ID_TESTKIT_CLI


@Factory(plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id=METTA_PLUGIN_ID_TESTKIT_PROVISIONER)
def metta_plugin_factory_provisioner_testkit(
        environment: Environment, instance_id: str = '',
        label: str = TESTKIT_PROVISIONER_CONFIG_LABEL, base: Any = TESTKIT_PROVISIONER_CONFIG_BASE):
    """Create an metta litmuschaos workload plugin."""
    return TestkitProvisionerPlugin(environment, instance_id, label=label, base=base)


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_PLUGIN_ID_TESTKIT_CLI)
def metta_plugin_factory_cli_testkit(environment: Environment, instance_id: str = ''):
    """Create an litmuschaos cli plugin."""
    return TestkitCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


def bootstrap():
    """METTA_testkit bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added
        to.

    """



"""

Kubernetes workload plugin.

This plugin uses the kube_api client to deploy workloads to kubernetes, from a deployment
definition from configuration.

"""

import logging
from typing import Any

import kubernetes

from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta.workload import WorkloadBase, WorkloadInstanceBase

from .kubeapi_client import METTA_PLUGIN_ID_KUBERNETES_CLIENT

logger = logging.getLogger('metta.contrib.kubernetes.workload.deployment')

KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL = 'kubernetes'
KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE = 'workload.deployment'

KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_NAMESPACE = "namespace"
KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY = "body"

METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD = 'metta_kubernetes_deployment'
""" workload plugin_id for the metta_kubernetes deployment plugin """


class KubernetesDeploymentWorkloadPlugin(WorkloadBase):
    """Metta workload plugin for Kubernetes workload."""

    def __init__(self, environment, instance_id,
                 label: str = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL,
                 base: Any = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE):
        """Run the super constructor but also set class properties.

        This implements the args part of the client interface.

        Here we expect to receive a path to a KUBECONFIG file with a context set
        and we create a Kubernetes client for use.  After that this can provide
        Core api clients as per the kubernetes SDK

        Parameters:
        -----------
        config_file (str): String path to the kubernetes config file to use

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """Create a workload instance from a set of fixtures.

        Parameters:
        ----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """
        try:
            client = fixtures.get_plugin(plugin_type=METTA_PLUGIN_TYPE_CLIENT,
                                         plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)
        except KeyError as err:
            raise NotImplementedError("Workload could not find the needed client: "
                                      f"{METTA_PLUGIN_ID_KUBERNETES_CLIENT}") from err

        workload_config = self.environment.config.load(self.config_label)

        name = workload_config.get([self.config_base,
                                    KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY,
                                    'metadata.name'],
                                   default='not-declared')
        namespace = workload_config.get(
            [self.config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_NAMESPACE])
        body = workload_config.get(
            [self.config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY])

        return KubernetesDeploymentWorkloadInstance(
            client, namespace, name, body)

    def info(self):
        """Return dict data about this plugin for introspection."""
        workload_config = self.environment.config.load(self.config_label)

        return {
            'workload': {
                'deployment': {
                    'namespace': workload_config.get(
                        [self.config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_NAMESPACE]),
                    'body': workload_config.get(
                        [self.config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY])
                },
                'required_fixtures': {
                    'kubernetes': {
                        'plugin_type': METTA_PLUGIN_TYPE_CLIENT,
                        'plugin_id': 'metta_kubernetes'
                    }
                }
            }
        }


class KubernetesDeploymentWorkloadInstance(WorkloadInstanceBase):
    """Workload plugin instance, for managing individual deployment runs."""

    def __init__(self, client, namespace, name, body):
        """Set initial workload instance state."""
        self.client = client
        self.namespace = namespace
        self.name = name
        self.body = body

        self.deployment = None

        self.read()

    def read(self):
        """Retrieve the deployment job."""
        if self.deployment is None:
            apps_v1 = self.client.get_api('AppsV1Api')

            try:
                self.deployment = apps_v1.read_namespaced_deployment(
                    self.name, self.namespace)
            except kubernetes.client.rest.ApiException:
                self.deployment = None

        return self.deployment

    def apply(self):
        """Run the workload."""
        apps_v1 = self.client.get_api('AppsV1Api')
        self.deployment = apps_v1.create_namespaced_deployment(
            body=self.body, namespace=self.namespace)

        return self.deployment

    def destroy(self):
        """Destroy any created resources."""
        body = kubernetes.client.V1DeleteOptions(
            propagation_policy='Foreground',
            grace_period_seconds=5)

        apps_v1 = self.client.get_api('AppsV1Api')
        status = apps_v1.delete_namespaced_deployment(
            name=self.name, namespace=self.namespace, body=body)

        self.deployment = None

        return status



"""

Metta provisioner plugin for testkit.

"""
import logging
import os
from typing import Any, Dict

import yaml

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
from configerus.validator import ValidationError

from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.provisioner import ProvisionerBase

from .testkit import TestkitClient, TESTKITCLIENT_CLI_CONFIG_FILE_DEFAULT

logger = logging.getLogger('testkit.provisioner')

METTA_PLUGIN_ID_TESTKIT_PROVISIONER = 'metta_testkit'
""" Metta plugin id for testkit provisioner plugins """

TESTKIT_PROVISIONER_CONFIG_LABEL = 'testkit'
TESTKIT_PROVISIONER_CONFIG_BASE = LOADED_KEY_ROOT
""" configerus config label/base-key for loading testkit provisioner config """

TESTKIT_CONFIG_KEY_SYSTEMNAME = 'system_name'
""" config key to find the system name """
TESTKIT_CONFIG_KEY_CREATE_OPTIONS = 'options.create'
""" config key to find where the testkit create options """
TESTKIT_CONFIG_KEY_CONFIG = 'config'
""" config key to find where to put the testkit config file data/contents """
TESTKIT_CONFIG_KEY_INSTANCES = 'instances'
""" config key to find what instances to spec out (build) """
TESTKIT_CONFIG_KEY_CONFIGFILE = 'config_file'
""" config key to find where to put the testkit config file """
TESTKIT_CONFIG_DEFAULT_CONFIGFILE = TESTKITCLIENT_CLI_CONFIG_FILE_DEFAULT
""" default value for where to put the testkit config file """

METTA_TESTKIT_CONFIG_VALIDATE_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        'mcr': {'type': 'object'},
        'mke': {'type': 'object'},
        'msr': {'type': 'object'}
    },
    'required': []
}
""" Validation jsonschema for testkit configuration for testkit yaml files """
METTA_TESTKIT_CONFIG_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: METTA_TESTKIT_CONFIG_VALIDATE_JSONSCHEMA
}
""" configerus jsonschema validation target for testkit config file """

METTA_TESTKIT_PROVISIONER_CONFIG_VALIDATE_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        'opts': {'type': 'object'},
        'config': METTA_TESTKIT_CONFIG_VALIDATE_JSONSCHEMA
    },
    'required': []
}
""" Validation jsonschema for provisioner configuration """
METTA_TESTKIT_PROVISIONER_CONFIG_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: METTA_TESTKIT_PROVISIONER_CONFIG_VALIDATE_JSONSCHEMA
}
""" configerus jsonschema validation target for the provisioner plugin """


# pylint: disable=too-many-instance-attributes
class TestkitProvisionerPlugin(ProvisionerBase):
    """Testkit provisioner plugin.

    Provisioner plugin that allows control of and interaction with a testkit
    cluster.

    ## Requirements

    1. this plugin uses subprocess to call a testkit binary, so you have to
       install testkit in the environment

    ## Usage

    @TODO

    """

    def __init__(self, environment, instance_id, label: str = TESTKIT_PROVISIONER_CONFIG_LABEL,
                 base: Any = TESTKIT_PROVISIONER_CONFIG_BASE):
        """Initialize Testkit provisioner.

        Parameters:
        -----------
        environment (Environment) : metta environment object that this plugin
            is attached.
        instance_id (str) : label for this plugin instances.

        label (str) : config load label for plugin configuration.
        base (str) : config base for loaded config for plugin configuration.

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        self.fixtures = Fixtures()
        """ Keep a collection of all fixtures created by this plugin """

        logger.info("Preparing Testkit setting")

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base

        testkit_config = self.environment.config.load(self.config_label, force_reload=True)
        """ load the plugin configuration so we can retrieve options """

        self.system_name = testkit_config.get([self.config_base, TESTKIT_CONFIG_KEY_SYSTEMNAME])
        """ hat will testkit call the system """

        try:
            testkit_config = self.environment.config.load(self.config_label)
            """ loaded plugin configuration label """
        except KeyError as err:
            raise ValueError("Testkit plugin configuration did not have any config") from err

        # instances = testkit_config.get([self.config_base, TESTKIT_CONFIG_KEY_INSTANCES])
        # """ what instances to create """

        self.config_file = testkit_config.get([self.config_base, TESTKIT_CONFIG_KEY_CONFIGFILE],
                                              default=TESTKIT_CONFIG_DEFAULT_CONFIGFILE)
        """ config_file value from plugin configuration """

        self.testkit = TestkitClient(config_file=self.config_file)
        """ testkit client object """

    # the deep argument is a standard for the info hook
    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Get info about a provisioner plugin."""
        plugin = self
        client = self.testkit
        testkit_config = self.environment.config.load(self.config_label)

        info = {
            'plugin': {
                'config_label': plugin.config_label,
                'config_base': plugin.config_base,
                'system_name': self.system_name
            },
            'client': {
                'config_file': client.config_file,
                'working_dir': client.working_dir,
                'bin': client.bin,
                'version': client.version()
            },
            # 'instances': testkit_config.get([self.config_base, TESTKIT_CONFIG_KEY_INSTANCES]),
            'config': testkit_config.get([self.config_base, TESTKIT_CONFIG_KEY_CONFIG])
        }

        return info

    def prepare(self):
        """Prepare any needed resources.

        We don't create the testkit file here so that it is created as late as
        possible.  This allows more options for dynamic config sources in the
        testkit config.

        """

    def apply(self):
        """Create the testkit yaml file and run testkit to create a cluster."""
        self._write_config_file()

        testkit_config = self.environment.config.load(self.config_label, force_reload=True)
        """ load the plugin configuration so we can retrieve options """
        opts = testkit_config.get([self.config_base, TESTKIT_CONFIG_KEY_CREATE_OPTIONS], default={})
        """ retrieve testkit client options from config """
        opt_list = []
        for key, value in opts.items():
            opt_list.append(f'--{key}={value}')

        # run the testkit client command to provisioner the cluster
        self.testkit.create(opts=opt_list)

    def destroy(self):
        """Destroy any created resources."""
        # run the testkit client command to provisioner the cluster
        self.testkit.system_rm(self.system_name)
        self._rm_config_file()

    def _write_config_file(self):
        """Write the config file for testkit."""
        try:
            # load all of the testkit configuration, force a reload to get up to date contents
            testkit_config = self.environment.config.load(self.config_label, force_reload=True)
            config = testkit_config.get([self.config_base, TESTKIT_CONFIG_KEY_CONFIG],
                                        validator=METTA_TESTKIT_CONFIG_VALIDATE_TARGET)
            """ config source of launchpad yaml """
        except KeyError as err:
            raise ValueError("Could not find launchpad configuration from config.") from err
        except ValidationError as err:
            raise ValueError("Launchpad config failed validation") from err

        # write the configto our yaml file target (creating the path)
        os.makedirs(os.path.dirname(os.path.realpath(self.config_file)), exist_ok=True)
        with open(os.path.realpath(self.config_file), 'w') as file:
            yaml.dump(config, file)

    def _rm_config_file(self):
        """Remove the written config file."""
        if os.path.isfile(self.config_file):
            os.remove(self.config_file)
